import contextlib
import json
import os
import subprocess
import sys
import time
import warnings
from collections import Counter, defaultdict
from collections.abc import Mapping
from textwrap import dedent
from unittest import mock

import pytest

import sentry_sdk
from sentry_sdk import (
    Client,
    Hub,
    add_breadcrumb,
    capture_event,
    capture_exception,
    capture_message,
    configure_scope,
    set_tag,
    start_transaction,
)
from sentry_sdk._compat import PY38
from sentry_sdk.consts import DEFAULT_MAX_BREADCRUMBS
from sentry_sdk.integrations.asyncio import AsyncioIntegration
from sentry_sdk.integrations.executing import ExecutingIntegration
from sentry_sdk.serializer import MAX_DATABAG_BREADTH
from sentry_sdk.spotlight import DEFAULT_SPOTLIGHT_URL
from sentry_sdk.transport import AsyncHttpTransport, Transport
from sentry_sdk.utils import capture_internal_exception

try:
    import gevent  # noqa: F401

    running_under_gevent = True
except ImportError:
    running_under_gevent = False

skip_under_gevent = pytest.mark.skipif(
    running_under_gevent, reason="Async tests not compatible with gevent"
)

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from collections.abc import Callable
    from typing import Any, Optional, Union

    from sentry_sdk._types import Event


maximum_python_312 = pytest.mark.skipif(
    sys.version_info > (3, 12),
    reason="Since Python 3.13, `FrameLocalsProxy` skips items of `locals()` that have non-`str` keys; this is a CPython implementation detail: https://github.com/python/cpython/blame/7b413952e817ae87bfda2ac85dd84d30a6ce743b/Objects/frameobject.c#L148",
)

PROXY_TESTCASES = [
    {
        "dsn": "http://foo@sentry.io/123",
        "env_http_proxy": None,
        "env_https_proxy": None,
        "arg_http_proxy": "http://localhost/123",
        "arg_https_proxy": None,
        "expected_proxy_scheme": "http",
    },
    {
        "dsn": "https://foo@sentry.io/123",
        "env_http_proxy": None,
        "env_https_proxy": None,
        "arg_http_proxy": "https://localhost/123",
        "arg_https_proxy": None,
        "expected_proxy_scheme": "https",
    },
    {
        "dsn": "http://foo@sentry.io/123",
        "env_http_proxy": None,
        "env_https_proxy": None,
        "arg_http_proxy": "http://localhost/123",
        "arg_https_proxy": "https://localhost/123",
        "expected_proxy_scheme": "http",
    },
    {
        "dsn": "https://foo@sentry.io/123",
        "env_http_proxy": None,
        "env_https_proxy": None,
        "arg_http_proxy": "http://localhost/123",
        "arg_https_proxy": "https://localhost/123",
        "expected_proxy_scheme": "https",
    },
    {
        "dsn": "https://foo@sentry.io/123",
        "env_http_proxy": None,
        "env_https_proxy": None,
        "arg_http_proxy": "http://localhost/123",
        "arg_https_proxy": None,
        "expected_proxy_scheme": "http",
    },
    {
        "dsn": "http://foo@sentry.io/123",
        "env_http_proxy": None,
        "env_https_proxy": None,
        "arg_http_proxy": None,
        "arg_https_proxy": None,
        "expected_proxy_scheme": None,
    },
    {
        "dsn": "http://foo@sentry.io/123",
        "env_http_proxy": "http://localhost/123",
        "env_https_proxy": None,
        "arg_http_proxy": None,
        "arg_https_proxy": None,
        "expected_proxy_scheme": "http",
    },
    {
        "dsn": "https://foo@sentry.io/123",
        "env_http_proxy": None,
        "env_https_proxy": "https://localhost/123",
        "arg_http_proxy": None,
        "arg_https_proxy": None,
        "expected_proxy_scheme": "https",
    },
    {
        "dsn": "https://foo@sentry.io/123",
        "env_http_proxy": "http://localhost/123",
        "env_https_proxy": None,
        "arg_http_proxy": None,
        "arg_https_proxy": None,
        "expected_proxy_scheme": "http",
    },
    {
        "dsn": "https://foo@sentry.io/123",
        "env_http_proxy": "http://localhost/123",
        "env_https_proxy": "https://localhost/123",
        "arg_http_proxy": "",
        "arg_https_proxy": "",
        "expected_proxy_scheme": None,
    },
    {
        "dsn": "https://foo@sentry.io/123",
        "env_http_proxy": "http://localhost/123",
        "env_https_proxy": "https://localhost/123",
        "arg_http_proxy": None,
        "arg_https_proxy": None,
        "expected_proxy_scheme": "https",
    },
    {
        "dsn": "https://foo@sentry.io/123",
        "env_http_proxy": "http://localhost/123",
        "env_https_proxy": None,
        "arg_http_proxy": None,
        "arg_https_proxy": None,
        "expected_proxy_scheme": "http",
    },
    {
        "dsn": "https://foo@sentry.io/123",
        "env_http_proxy": "http://localhost/123",
        "env_https_proxy": "https://localhost/123",
        "arg_http_proxy": None,
        "arg_https_proxy": "",
        "expected_proxy_scheme": "http",
    },
    {
        "dsn": "https://foo@sentry.io/123",
        "env_http_proxy": "http://localhost/123",
        "env_https_proxy": "https://localhost/123",
        "arg_http_proxy": "",
        "arg_https_proxy": None,
        "expected_proxy_scheme": "https",
    },
    {
        "dsn": "https://foo@sentry.io/123",
        "env_http_proxy": None,
        "env_https_proxy": "https://localhost/123",
        "arg_http_proxy": None,
        "arg_https_proxy": "",
        "expected_proxy_scheme": None,
    },
    {
        "dsn": "http://foo@sentry.io/123",
        "env_http_proxy": "http://localhost/123",
        "env_https_proxy": "https://localhost/123",
        "arg_http_proxy": None,
        "arg_https_proxy": None,
        "expected_proxy_scheme": "http",
    },
    # NO_PROXY testcases
    {
        "dsn": "http://foo@sentry.io/123",
        "env_http_proxy": "http://localhost/123",
        "env_https_proxy": None,
        "env_no_proxy": "sentry.io,example.com",
        "arg_http_proxy": None,
        "arg_https_proxy": None,
        "expected_proxy_scheme": None,
    },
    {
        "dsn": "https://foo@sentry.io/123",
        "env_http_proxy": None,
        "env_https_proxy": "https://localhost/123",
        "env_no_proxy": "example.com,sentry.io",
        "arg_http_proxy": None,
        "arg_https_proxy": None,
        "expected_proxy_scheme": None,
    },
    {
        "dsn": "http://foo@sentry.io/123",
        "env_http_proxy": None,
        "env_https_proxy": None,
        "env_no_proxy": "sentry.io,example.com",
        "arg_http_proxy": "http://localhost/123",
        "arg_https_proxy": None,
        "expected_proxy_scheme": "http",
    },
    {
        "dsn": "https://foo@sentry.io/123",
        "env_http_proxy": None,
        "env_https_proxy": None,
        "env_no_proxy": "sentry.io,example.com",
        "arg_http_proxy": None,
        "arg_https_proxy": "https://localhost/123",
        "expected_proxy_scheme": "https",
    },
    {
        "dsn": "https://foo@sentry.io/123",
        "env_http_proxy": None,
        "env_https_proxy": None,
        "env_no_proxy": "sentry.io,example.com",
        "arg_http_proxy": None,
        "arg_https_proxy": "https://localhost/123",
        "expected_proxy_scheme": "https",
        "arg_proxy_headers": {"Test-Header": "foo-bar"},
    },
]

SOCKS_PROXY_TESTCASES = [
    {
        "dsn": "https://foo@sentry.io/123",
        "arg_http_proxy": "http://localhost/123",
        "arg_https_proxy": None,
        "should_be_socks_proxy": False,
    },
    {
        "dsn": "https://foo@sentry.io/123",
        "arg_http_proxy": "socks4a://localhost/123",
        "arg_https_proxy": None,
        "should_be_socks_proxy": True,
    },
    {
        "dsn": "https://foo@sentry.io/123",
        "arg_http_proxy": "socks4://localhost/123",
        "arg_https_proxy": None,
        "should_be_socks_proxy": True,
    },
    {
        "dsn": "https://foo@sentry.io/123",
        "arg_http_proxy": "socks5h://localhost/123",
        "arg_https_proxy": None,
        "should_be_socks_proxy": True,
    },
    {
        "dsn": "https://foo@sentry.io/123",
        "arg_http_proxy": "socks5://localhost/123",
        "arg_https_proxy": None,
        "should_be_socks_proxy": True,
    },
    {
        "dsn": "https://foo@sentry.io/123",
        "arg_http_proxy": None,
        "arg_https_proxy": "socks4a://localhost/123",
        "should_be_socks_proxy": True,
    },
    {
        "dsn": "https://foo@sentry.io/123",
        "arg_http_proxy": None,
        "arg_https_proxy": "socks4://localhost/123",
        "should_be_socks_proxy": True,
    },
    {
        "dsn": "https://foo@sentry.io/123",
        "arg_http_proxy": None,
        "arg_https_proxy": "socks5h://localhost/123",
        "should_be_socks_proxy": True,
    },
    {
        "dsn": "https://foo@sentry.io/123",
        "arg_http_proxy": None,
        "arg_https_proxy": "socks5://localhost/123",
        "should_be_socks_proxy": True,
    },
]


class EnvelopeCapturedError(Exception):
    pass


class _TestTransport(Transport):
    def capture_envelope(self, envelope):
        raise EnvelopeCapturedError(envelope)


def test_transport_option(monkeypatch):
    if "SENTRY_DSN" in os.environ:
        monkeypatch.delenv("SENTRY_DSN")

    dsn = "https://foo@sentry.io/123"
    dsn2 = "https://bar@sentry.io/124"
    assert str(Client(dsn=dsn).dsn) == dsn
    assert Client().dsn is None

    monkeypatch.setenv("SENTRY_DSN", dsn)
    transport = _TestTransport({"dsn": dsn2})
    assert str(transport.parsed_dsn) == dsn2
    assert str(Client(transport=transport).dsn) == dsn


@pytest.mark.parametrize("testcase", PROXY_TESTCASES)
@pytest.mark.parametrize(
    "http2", [True, False] if sys.version_info >= (3, 8) else [False]
)
def test_proxy(monkeypatch, testcase, http2):
    if testcase["env_http_proxy"] is not None:
        monkeypatch.setenv("HTTP_PROXY", testcase["env_http_proxy"])
    if testcase["env_https_proxy"] is not None:
        monkeypatch.setenv("HTTPS_PROXY", testcase["env_https_proxy"])
    if testcase.get("env_no_proxy") is not None:
        monkeypatch.setenv("NO_PROXY", testcase["env_no_proxy"])

    kwargs = {}

    if http2:
        kwargs["_experiments"] = {"transport_http2": True}

    if testcase["arg_http_proxy"] is not None:
        kwargs["http_proxy"] = testcase["arg_http_proxy"]
    if testcase["arg_https_proxy"] is not None:
        kwargs["https_proxy"] = testcase["arg_https_proxy"]
    if testcase.get("arg_proxy_headers") is not None:
        kwargs["proxy_headers"] = testcase["arg_proxy_headers"]

    client = Client(testcase["dsn"], **kwargs)

    proxy = getattr(
        client.transport._pool,
        "proxy",
        getattr(client.transport._pool, "_proxy_url", None),
    )
    if testcase["expected_proxy_scheme"] is None:
        assert proxy is None
    else:
        scheme = (
            proxy.scheme.decode("ascii")
            if isinstance(proxy.scheme, bytes)
            else proxy.scheme
        )
        assert scheme == testcase["expected_proxy_scheme"]

        if testcase.get("arg_proxy_headers") is not None:
            proxy_headers = (
                dict(
                    (k.decode("ascii"), v.decode("ascii"))
                    for k, v in client.transport._pool._proxy_headers
                )
                if http2
                else client.transport._pool.proxy_headers
            )
            assert proxy_headers == testcase["arg_proxy_headers"]


@pytest.mark.parametrize("testcase", SOCKS_PROXY_TESTCASES)
@pytest.mark.parametrize(
    "http2", [True, False] if sys.version_info >= (3, 8) else [False]
)
def test_socks_proxy(testcase, http2):
    kwargs = {}

    if http2:
        kwargs["_experiments"] = {"transport_http2": True}

    if testcase["arg_http_proxy"] is not None:
        kwargs["http_proxy"] = testcase["arg_http_proxy"]
    if testcase["arg_https_proxy"] is not None:
        kwargs["https_proxy"] = testcase["arg_https_proxy"]

    client = Client(testcase["dsn"], **kwargs)
    assert ("socks" in str(type(client.transport._pool)).lower()) == testcase[
        "should_be_socks_proxy"
    ], (
        f"Expected {kwargs} to result in SOCKS == {testcase['should_be_socks_proxy']}"
        f"but got {str(type(client.transport._pool))}"
    )


def test_simple_transport(sentry_init):
    events = []
    sentry_init(transport=events.append)
    capture_message("Hello World!")
    assert events[0]["message"] == "Hello World!"


def test_ignore_errors(sentry_init, capture_events):
    sentry_init(ignore_errors=[ZeroDivisionError])
    events = capture_events()

    class MyDivisionError(ZeroDivisionError):
        pass

    def e(exc):
        try:
            raise exc
        except Exception:
            capture_exception()

    e(ZeroDivisionError())
    e(MyDivisionError())
    e(ValueError())

    assert len(events) == 1
    assert events[0]["exception"]["values"][0]["type"] == "ValueError"


def test_include_local_variables_enabled(sentry_init, capture_events):
    sentry_init(include_local_variables=True)
    events = capture_events()
    try:
        1 / 0
    except Exception:
        capture_exception()

    (event,) = events

    assert all(
        frame["vars"]
        for frame in event["exception"]["values"][0]["stacktrace"]["frames"]
    )


def test_include_local_variables_disabled(sentry_init, capture_events):
    sentry_init(include_local_variables=False)
    events = capture_events()
    try:
        1 / 0
    except Exception:
        capture_exception()

    (event,) = events

    assert all(
        "vars" not in frame
        for frame in event["exception"]["values"][0]["stacktrace"]["frames"]
    )


def test_include_source_context_enabled(sentry_init, capture_events):
    sentry_init(include_source_context=True)
    events = capture_events()
    try:
        1 / 0
    except Exception:
        capture_exception()

    (event,) = events

    frame = event["exception"]["values"][0]["stacktrace"]["frames"][0]
    assert "post_context" in frame
    assert "pre_context" in frame
    assert "context_line" in frame


def test_include_source_context_disabled(sentry_init, capture_events):
    sentry_init(include_source_context=False)
    events = capture_events()
    try:
        1 / 0
    except Exception:
        capture_exception()

    (event,) = events

    frame = event["exception"]["values"][0]["stacktrace"]["frames"][0]
    assert "post_context" not in frame
    assert "pre_context" not in frame
    assert "context_line" not in frame


@pytest.mark.parametrize("integrations", [[], [ExecutingIntegration()]])
def test_function_names(sentry_init, capture_events, integrations):
    sentry_init(integrations=integrations)
    events = capture_events()

    def foo():
        try:
            bar()
        except Exception:
            capture_exception()

    def bar():
        1 / 0

    foo()

    (event,) = events
    (thread,) = event["exception"]["values"]
    functions = [x["function"] for x in thread["stacktrace"]["frames"]]

    if integrations:
        assert functions == [
            "test_function_names.<locals>.foo",
            "test_function_names.<locals>.bar",
        ]
    else:
        assert functions == ["foo", "bar"]


def test_attach_stacktrace_enabled(sentry_init, capture_events):
    sentry_init(attach_stacktrace=True)
    events = capture_events()

    def foo():
        bar()

    def bar():
        capture_message("HI")

    foo()

    (event,) = events
    (thread,) = event["threads"]["values"]
    functions = [x["function"] for x in thread["stacktrace"]["frames"]]

    assert functions[-2:] == ["foo", "bar"]


def test_attach_stacktrace_enabled_no_locals(sentry_init, capture_events):
    sentry_init(attach_stacktrace=True, include_local_variables=False)
    events = capture_events()

    def foo():
        bar()

    def bar():
        capture_message("HI")

    foo()

    (event,) = events
    (thread,) = event["threads"]["values"]
    local_vars = [x.get("vars") for x in thread["stacktrace"]["frames"]]
    assert local_vars[-2:] == [None, None]


def test_attach_stacktrace_in_app(sentry_init, capture_events):
    sentry_init(attach_stacktrace=True, in_app_exclude=["_pytest"])
    events = capture_events()

    capture_message("hi")

    (event,) = events
    (thread,) = event["threads"]["values"]
    frames = thread["stacktrace"]["frames"]
    pytest_frames = [f for f in frames if f["module"].startswith("_pytest")]
    assert pytest_frames
    assert all(f["in_app"] is False for f in pytest_frames)


def test_attach_stacktrace_disabled(sentry_init, capture_events):
    sentry_init(attach_stacktrace=False)
    events = capture_events()
    capture_message("HI")

    (event,) = events
    assert "threads" not in event


def test_attach_stacktrace_transaction(sentry_init, capture_events):
    sentry_init(traces_sample_rate=1.0, attach_stacktrace=True)
    events = capture_events()
    with start_transaction(name="transaction"):
        pass
    (event,) = events
    assert "threads" not in event


def test_capture_event_works(sentry_init):
    sentry_init(transport=_TestTransport())
    pytest.raises(EnvelopeCapturedError, lambda: capture_event({}))
    pytest.raises(EnvelopeCapturedError, lambda: capture_event({}))


@pytest.mark.parametrize("num_messages", [10, 20])
@pytest.mark.parametrize(
    "http2", [True, False] if sys.version_info >= (3, 8) else [False]
)
def test_atexit(tmpdir, monkeypatch, num_messages, http2):
    if http2:
        options = '_experiments={"transport_http2": True}'
        transport = "Http2Transport"
    else:
        options = ""
        transport = "HttpTransport"

    app = tmpdir.join("app.py")
    app.write(
        dedent(
            """
    import time
    from sentry_sdk import init, transport, capture_message

    def capture_envelope(self, envelope):
        time.sleep(0.1)
        event = envelope.get_event() or dict()
        message = event.get("message", "")
        print(message)

    transport.{transport}.capture_envelope = capture_envelope
    init("http://foobar@localhost/123", shutdown_timeout={num_messages}, {options})

    for _ in range({num_messages}):
        capture_message("HI")
    """.format(transport=transport, options=options, num_messages=num_messages)
        )
    )

    start = time.time()
    output = subprocess.check_output([sys.executable, str(app)])
    end = time.time()

    # Each message takes at least 0.1 seconds to process
    assert int(end - start) >= num_messages / 10

    assert output.count(b"HI") == num_messages


def test_configure_scope_available(
    sentry_init, request, monkeypatch, suppress_deprecation_warnings
):
    """
    Test that scope is configured if client is configured

    This test can be removed once configure_scope and the Hub are removed.
    """
    sentry_init()

    with configure_scope() as scope:
        assert scope is Hub.current.scope
        scope.set_tag("foo", "bar")

    calls = []

    def callback(scope):
        calls.append(scope)
        scope.set_tag("foo", "bar")

    assert configure_scope(callback) is None
    assert len(calls) == 1
    assert calls[0] is Hub.current.scope


@pytest.mark.tests_internal_exceptions
def test_client_debug_option_enabled(sentry_init, caplog):
    sentry_init(debug=True)

    capture_internal_exception((ValueError, ValueError("OK"), None))
    assert "OK" in caplog.text


@pytest.mark.tests_internal_exceptions
@pytest.mark.parametrize("with_client", (True, False))
def test_client_debug_option_disabled(with_client, sentry_init, caplog):
    if with_client:
        sentry_init()

    capture_internal_exception((ValueError, ValueError("OK"), None))
    assert "OK" not in caplog.text


@pytest.mark.skip(
    reason="New behavior in SDK 2.0: You have a scope before init and add data to it."
)
def test_scope_initialized_before_client(sentry_init, capture_events):
    """
    This is a consequence of how configure_scope() works. We must
    make `configure_scope()` a noop if no client is configured. Even
    if the user later configures a client: We don't know that.
    """
    with configure_scope() as scope:
        scope.set_tag("foo", 42)

    sentry_init()

    events = capture_events()
    capture_message("hi")
    (event,) = events

    assert "tags" not in event


def test_weird_chars(sentry_init, capture_events):
    sentry_init()
    events = capture_events()
    capture_message("föö".encode("latin1"))
    (event,) = events
    assert json.loads(json.dumps(event)) == event


def test_nan(sentry_init, capture_events):
    sentry_init()
    events = capture_events()

    try:
        # should_repr_strings=False
        set_tag("mynan", float("nan"))

        # should_repr_strings=True
        nan = float("nan")  # noqa
        1 / 0
    except Exception:
        capture_exception()

    (event,) = events
    frames = event["exception"]["values"][0]["stacktrace"]["frames"]
    (frame,) = frames
    assert frame["vars"]["nan"] == "nan"
    assert event["tags"]["mynan"] == "nan"


def test_cyclic_frame_vars(sentry_init, capture_events):
    sentry_init()
    events = capture_events()

    try:
        a = {}
        a["a"] = a
        1 / 0
    except Exception:
        capture_exception()

    (event,) = events
    assert event["exception"]["values"][0]["stacktrace"]["frames"][0]["vars"]["a"] == {
        "a": "<cyclic>"
    }


def test_cyclic_data(sentry_init, capture_events):
    sentry_init()
    events = capture_events()

    data = {}
    data["is_cyclic"] = data

    other_data = ""
    data["not_cyclic"] = other_data
    data["not_cyclic2"] = other_data
    sentry_sdk.get_isolation_scope().set_extra("foo", data)

    capture_message("hi")
    (event,) = events

    data = event["extra"]["foo"]
    assert data == {"not_cyclic2": "", "not_cyclic": "", "is_cyclic": "<cyclic>"}


def test_databag_depth_stripping(sentry_init, capture_events):
    sentry_init()
    events = capture_events()

    value = ["a"]
    for _ in range(100000):
        value = [value]

    del events[:]
    try:
        a = value  # noqa
        1 / 0
    except Exception:
        capture_exception()

    (event,) = events

    stacktrace_frame = event["exception"]["values"][0]["stacktrace"]["frames"][0]
    a_var = stacktrace_frame["vars"]["a"]

    assert type(a_var) == list
    assert len(a_var) == 1 and type(a_var[0]) == list

    first_level_list = a_var[0]
    assert type(first_level_list) == list
    assert len(first_level_list) == 1

    second_level_list = first_level_list[0]
    assert type(second_level_list) == list
    assert len(second_level_list) == 1

    third_level_list = second_level_list[0]
    assert type(third_level_list) == list
    assert len(third_level_list) == 1

    inner_value_repr = third_level_list[0]
    assert type(inner_value_repr) == str


def test_databag_string_stripping(sentry_init, capture_events):
    sentry_init(max_value_length=1024)
    events = capture_events()

    del events[:]
    try:
        a = "A" * 10240  # noqa
        1 / 0
    except Exception:
        capture_exception()

    (event,) = events

    assert len(json.dumps(event)) < 10240


def test_databag_breadth_stripping(sentry_init, capture_events):
    sentry_init()
    events = capture_events()

    del events[:]
    try:
        a = ["a"] * 1000000  # noqa
        1 / 0
    except Exception:
        capture_exception()

    (event,) = events

    assert (
        len(event["exception"]["values"][0]["stacktrace"]["frames"][0]["vars"]["a"])
        == MAX_DATABAG_BREADTH
    )
    assert len(json.dumps(event)) < 10000


def test_chained_exceptions(sentry_init, capture_events):
    sentry_init()
    events = capture_events()

    try:
        try:
            raise ValueError()
        except Exception:
            1 / 0
    except Exception:
        capture_exception()

    (event,) = events

    e1, e2 = event["exception"]["values"]

    # This is the order all other SDKs send chained exceptions in. Including
    # Raven-Python.

    assert e1["type"] == "ValueError"
    assert e2["type"] == "ZeroDivisionError"


@pytest.mark.tests_internal_exceptions
def test_broken_mapping(sentry_init, capture_events):
    sentry_init()
    events = capture_events()

    class C(Mapping):
        def broken(self, *args, **kwargs):
            raise Exception("broken")

        __getitem__ = broken
        __setitem__ = broken
        __delitem__ = broken
        __iter__ = broken
        __len__ = broken

        def __repr__(self):
            return "broken"

    try:
        a = C()  # noqa
        1 / 0
    except Exception:
        capture_exception()

    (event,) = events
    assert (
        event["exception"]["values"][0]["stacktrace"]["frames"][0]["vars"]["a"]
        == "<failed to serialize, use init(debug=True) to see error logs>"
    )


def test_mapping_sends_exception(sentry_init, capture_events):
    sentry_init()
    events = capture_events()

    class C(Mapping):
        def __iter__(self):
            try:
                1 / 0
            except ZeroDivisionError:
                capture_exception()
            yield "hi"

        def __len__(self):
            """List length"""
            return 1

        def __getitem__(self, ii):
            """Get a list item"""
            if ii == "hi":
                return "hi"

            raise KeyError()

    try:
        a = C()  # noqa
        1 / 0
    except Exception:
        capture_exception()

    (event,) = events

    assert event["exception"]["values"][0]["stacktrace"]["frames"][0]["vars"]["a"] == {
        "hi": "'hi'"
    }


def test_object_sends_exception(sentry_init, capture_events):
    sentry_init()
    events = capture_events()

    class C:
        def __repr__(self):
            try:
                1 / 0
            except ZeroDivisionError:
                capture_exception()
            return "hi, i am a repr"

    try:
        a = C()  # noqa
        1 / 0
    except Exception:
        capture_exception()

    (event,) = events

    assert (
        event["exception"]["values"][0]["stacktrace"]["frames"][0]["vars"]["a"]
        == "hi, i am a repr"
    )


def test_errno_errors(sentry_init, capture_events):
    sentry_init()
    events = capture_events()

    class FooError(Exception):
        errno = 69

    capture_exception(FooError())

    (event,) = events

    (exception,) = event["exception"]["values"]
    assert exception["mechanism"]["meta"]["errno"]["number"] == 69


@maximum_python_312
def test_non_string_variables(sentry_init, capture_events):
    """There is some extremely terrible code in the wild that
    inserts non-strings as variable names into `locals()`."""

    sentry_init()
    events = capture_events()

    try:
        locals()[42] = True
        1 / 0
    except ZeroDivisionError:
        capture_exception()

    (event,) = events

    (exception,) = event["exception"]["values"]
    assert exception["type"] == "ZeroDivisionError"
    (frame,) = exception["stacktrace"]["frames"]
    assert frame["vars"]["42"] == "True"


def test_dict_changed_during_iteration(sentry_init, capture_events):
    """
    Some versions of Bottle modify the WSGI environment inside of this __repr__
    impl: https://github.com/bottlepy/bottle/blob/0.12.16/bottle.py#L1386

    See https://github.com/getsentry/sentry-python/pull/298 for discussion
    """
    sentry_init(send_default_pii=True)
    events = capture_events()

    class TooSmartClass:
        def __init__(self, environ):
            self.environ = environ

        def __repr__(self):
            if "my_representation" in self.environ:
                return self.environ["my_representation"]

            self.environ["my_representation"] = "<This is me>"
            return self.environ["my_representation"]

    try:
        environ = {}
        environ["a"] = TooSmartClass(environ)
        1 / 0
    except ZeroDivisionError:
        capture_exception()

    (event,) = events
    (exception,) = event["exception"]["values"]
    (frame,) = exception["stacktrace"]["frames"]
    assert frame["vars"]["environ"] == {"a": "<This is me>"}


def test_custom_repr_on_vars(sentry_init, capture_events):
    class Foo:
        pass

    class Fail:
        pass

    def custom_repr(value):
        if isinstance(value, Foo):
            return "custom repr"
        elif isinstance(value, Fail):
            raise ValueError("oops")
        else:
            return None

    sentry_init(custom_repr=custom_repr)
    events = capture_events()

    try:
        my_vars = {"foo": Foo(), "fail": Fail(), "normal": 42}
        1 / 0
    except ZeroDivisionError:
        capture_exception()

    (event,) = events
    (exception,) = event["exception"]["values"]
    (frame,) = exception["stacktrace"]["frames"]
    my_vars = frame["vars"]["my_vars"]
    assert my_vars["foo"] == "custom repr"
    assert my_vars["normal"] == "42"
    assert "Fail object" in my_vars["fail"]


@pytest.mark.parametrize(
    "dsn",
    [
        "http://894b7d594095440f8dfea9b300e6f572@localhost:8000/2",
        "http://894b7d594095440f8dfea9b300e6f572@localhost:8000/2",
    ],
)
def test_init_string_types(dsn, sentry_init):
    # Allow unicode strings on Python 3 and both on Python 2 (due to
    # unicode_literals)
    #
    # Supporting bytes on Python 3 is not really wrong but probably would be
    # extra code
    sentry_init(dsn)
    assert (
        sentry_sdk.get_client().dsn
        == "http://894b7d594095440f8dfea9b300e6f572@localhost:8000/2"
    )


@pytest.mark.parametrize(
    "sdk_options, expected_breadcrumbs",
    [({}, DEFAULT_MAX_BREADCRUMBS), ({"max_breadcrumbs": 50}, 50)],
)
def test_max_breadcrumbs_option(
    sentry_init, capture_events, sdk_options, expected_breadcrumbs
):
    sentry_init(sdk_options)
    events = capture_events()

    for _ in range(1231):
        add_breadcrumb({"type": "sourdough"})

    capture_message("dogs are great")

    assert len(events[0]["breadcrumbs"]["values"]) == expected_breadcrumbs


def test_multiple_positional_args(sentry_init):
    with pytest.raises(TypeError) as exinfo:
        sentry_init(1, None)
    assert "Only single positional argument is expected" in str(exinfo.value)


def test_max_value_length_option(sentry_init, capture_events):
    sentry_init(max_value_length=2024)
    events = capture_events()

    capture_message("a" * (3024))

    assert len(events[0]["message"]) == 2024


@pytest.mark.parametrize(
    "client_option,env_var_value,debug_output_expected",
    [
        (None, "", False),
        (None, "t", True),
        (None, "1", True),
        (None, "True", True),
        (None, "true", True),
        (None, "f", False),
        (None, "0", False),
        (None, "False", False),
        (None, "false", False),
        (None, "xxx", False),
        (True, "", True),
        (True, "t", True),
        (True, "1", True),
        (True, "True", True),
        (True, "true", True),
        (True, "f", True),
        (True, "0", True),
        (True, "False", True),
        (True, "false", True),
        (True, "xxx", True),
        (False, "", False),
        (False, "t", False),
        (False, "1", False),
        (False, "True", False),
        (False, "true", False),
        (False, "f", False),
        (False, "0", False),
        (False, "False", False),
        (False, "false", False),
        (False, "xxx", False),
    ],
)
@pytest.mark.tests_internal_exceptions
def test_debug_option(
    sentry_init,
    monkeypatch,
    caplog,
    client_option,
    env_var_value,
    debug_output_expected,
):
    monkeypatch.setenv("SENTRY_DEBUG", env_var_value)

    if client_option is None:
        sentry_init()
    else:
        sentry_init(debug=client_option)

    capture_internal_exception((ValueError, ValueError("something is wrong"), None))
    if debug_output_expected:
        assert "something is wrong" in caplog.text
    else:
        assert "something is wrong" not in caplog.text


@pytest.mark.parametrize(
    "client_option,env_var_value,spotlight_url_expected",
    [
        (None, None, None),
        (None, "", None),
        (None, "F", None),
        (False, None, None),
        (False, "", None),
        (False, "t", None),
        (None, "t", DEFAULT_SPOTLIGHT_URL),
        (None, "1", DEFAULT_SPOTLIGHT_URL),
        (True, None, DEFAULT_SPOTLIGHT_URL),
        # Per spec: spotlight=True + env URL -> use env URL
        (True, "http://localhost:8080/slurp", "http://localhost:8080/slurp"),
        ("http://localhost:8080/slurp", "f", "http://localhost:8080/slurp"),
        (None, "http://localhost:8080/slurp", "http://localhost:8080/slurp"),
    ],
)
def test_spotlight_option(
    sentry_init,
    monkeypatch,
    client_option,
    env_var_value,
    spotlight_url_expected,
):
    if env_var_value is None:
        monkeypatch.delenv("SENTRY_SPOTLIGHT", raising=False)
    else:
        monkeypatch.setenv("SENTRY_SPOTLIGHT", env_var_value)

    if client_option is None:
        sentry_init()
    else:
        sentry_init(spotlight=client_option)

    client = sentry_sdk.get_client()
    url = client.spotlight.url if client.spotlight else None
    assert url == spotlight_url_expected, (
        f"With config {client_option} and env {env_var_value}"
    )


class IssuesSamplerTestConfig:
    def __init__(
        self,
        expected_events: int,
        sampler_function: "Optional[Callable[[Event], Union[float, bool]]]" = None,
        sample_rate: "Optional[float]" = None,
        exception_to_raise: "type[Exception]" = Exception,
    ) -> None:
        self.sampler_function_mock = (
            None
            if sampler_function is None
            else mock.MagicMock(side_effect=sampler_function)
        )
        self.expected_events = expected_events
        self.sample_rate = sample_rate
        self.exception_to_raise = exception_to_raise

    def init_sdk(self, sentry_init: "Callable[[*Any], None]") -> None:
        sentry_init(
            error_sampler=self.sampler_function_mock, sample_rate=self.sample_rate
        )

    def raise_exception(self) -> None:
        raise self.exception_to_raise()


@mock.patch("sentry_sdk.client.random.random", return_value=0.618)
@pytest.mark.parametrize(
    "test_config",
    (
        # Baseline test with error_sampler only, both floats and bools
        IssuesSamplerTestConfig(sampler_function=lambda *_: 1.0, expected_events=1),
        IssuesSamplerTestConfig(sampler_function=lambda *_: 0.7, expected_events=1),
        IssuesSamplerTestConfig(sampler_function=lambda *_: 0.6, expected_events=0),
        IssuesSamplerTestConfig(sampler_function=lambda *_: 0.0, expected_events=0),
        IssuesSamplerTestConfig(sampler_function=lambda *_: True, expected_events=1),
        IssuesSamplerTestConfig(sampler_function=lambda *_: False, expected_events=0),
        # Baseline test with sample_rate only
        IssuesSamplerTestConfig(sample_rate=1.0, expected_events=1),
        IssuesSamplerTestConfig(sample_rate=0.7, expected_events=1),
        IssuesSamplerTestConfig(sample_rate=0.6, expected_events=0),
        IssuesSamplerTestConfig(sample_rate=0.0, expected_events=0),
        # error_sampler takes precedence over sample_rate
        IssuesSamplerTestConfig(
            sampler_function=lambda *_: 1.0, sample_rate=0.0, expected_events=1
        ),
        IssuesSamplerTestConfig(
            sampler_function=lambda *_: 0.0, sample_rate=1.0, expected_events=0
        ),
        # Different sample rates based on exception, retrieved both from event and hint
        IssuesSamplerTestConfig(
            sampler_function=lambda event, _: {
                "ZeroDivisionError": 1.0,
                "AttributeError": 0.0,
            }[event["exception"]["values"][0]["type"]],
            exception_to_raise=ZeroDivisionError,
            expected_events=1,
        ),
        IssuesSamplerTestConfig(
            sampler_function=lambda event, _: {
                "ZeroDivisionError": 1.0,
                "AttributeError": 0.0,
            }[event["exception"]["values"][0]["type"]],
            exception_to_raise=AttributeError,
            expected_events=0,
        ),
        IssuesSamplerTestConfig(
            sampler_function=lambda _, hint: {
                ZeroDivisionError: 1.0,
                AttributeError: 0.0,
            }[hint["exc_info"][0]],
            exception_to_raise=ZeroDivisionError,
            expected_events=1,
        ),
        IssuesSamplerTestConfig(
            sampler_function=lambda _, hint: {
                ZeroDivisionError: 1.0,
                AttributeError: 0.0,
            }[hint["exc_info"][0]],
            exception_to_raise=AttributeError,
            expected_events=0,
        ),
        # If sampler returns invalid value, we should still send the event
        IssuesSamplerTestConfig(
            sampler_function=lambda *_: (
                "This is an invalid return value for the sampler"
            ),
            expected_events=1,
        ),
    ),
)
def test_error_sampler(_, sentry_init, capture_events, test_config):
    test_config.init_sdk(sentry_init)

    events = capture_events()

    try:
        test_config.raise_exception()
    except Exception:
        capture_exception()

    assert len(events) == test_config.expected_events

    if test_config.sampler_function_mock is not None:
        assert test_config.sampler_function_mock.call_count == 1

        # Ensure two arguments (the event and hint) were passed to the sampler function
        assert len(test_config.sampler_function_mock.call_args[0]) == 2


@pytest.mark.parametrize(
    "opt,missing_flags",
    [
        # lazy mode with enable-threads, no warning
        [{"enable-threads": True, "lazy-apps": True}, []],
        [{"enable-threads": "true", "lazy-apps": b"1"}, []],
        # preforking mode with enable-threads and py-call-uwsgi-fork-hooks, no warning
        [{"enable-threads": True, "py-call-uwsgi-fork-hooks": True}, []],
        [{"enable-threads": b"true", "py-call-uwsgi-fork-hooks": b"on"}, []],
        # lazy mode, no enable-threads, warning
        [{"lazy-apps": True}, ["--enable-threads"]],
        [{"enable-threads": b"false", "lazy-apps": True}, ["--enable-threads"]],
        [{"enable-threads": b"0", "lazy": True}, ["--enable-threads"]],
        # preforking mode, no enable-threads or py-call-uwsgi-fork-hooks, warning
        [{}, ["--enable-threads", "--py-call-uwsgi-fork-hooks"]],
        [{"processes": b"2"}, ["--enable-threads", "--py-call-uwsgi-fork-hooks"]],
        [{"enable-threads": True}, ["--py-call-uwsgi-fork-hooks"]],
        [{"enable-threads": b"1"}, ["--py-call-uwsgi-fork-hooks"]],
        [
            {"enable-threads": b"false"},
            ["--enable-threads", "--py-call-uwsgi-fork-hooks"],
        ],
        [{"py-call-uwsgi-fork-hooks": True}, ["--enable-threads"]],
    ],
)
def test_uwsgi_warnings(sentry_init, recwarn, opt, missing_flags):
    uwsgi = mock.MagicMock()
    uwsgi.opt = opt
    with mock.patch.dict("sys.modules", uwsgi=uwsgi):
        sentry_init(profiles_sample_rate=1.0)
        if missing_flags:
            assert len(recwarn) == 1
            record = recwarn.pop()
            for flag in missing_flags:
                assert flag in str(record.message)
        else:
            assert not recwarn


class TestSpanClientReports:
    """
    Tests for client reports related to spans.
    """

    __test__ = False

    @staticmethod
    def span_dropper(spans_to_drop):
        """
        Returns a function that can be used to drop spans from an event.
        """

        def drop_spans(event, _):
            event["spans"] = event["spans"][spans_to_drop:]
            return event

        return drop_spans

    @staticmethod
    def mock_transaction_event(span_count):
        """
        Returns a mock transaction event with the given number of spans.
        """

        return defaultdict(
            mock.MagicMock,
            type="transaction",
            spans=[mock.MagicMock() for _ in range(span_count)],
        )

    def __init__(self, span_count):
        """Configures a test case with the number of spans dropped and whether the transaction was dropped."""
        self.span_count = span_count
        self.expected_record_lost_event_calls = Counter()
        self.before_send = lambda event, _: event
        self.event_processor = lambda event, _: event

    def _update_resulting_calls(self, reason, drops_transactions=0, drops_spans=0):
        """
        Updates the expected calls with the given resulting calls.
        """
        if drops_transactions > 0:
            self.expected_record_lost_event_calls[
                (reason, "transaction", None, drops_transactions)
            ] += 1

        if drops_spans > 0:
            self.expected_record_lost_event_calls[
                (reason, "span", None, drops_spans)
            ] += 1

    def with_before_send(
        self,
        before_send,
        *,
        drops_transactions=0,
        drops_spans=0,
    ):
        self.before_send = before_send
        self._update_resulting_calls(
            "before_send",
            drops_transactions,
            drops_spans,
        )

        return self

    def with_event_processor(
        self,
        event_processor,
        *,
        drops_transactions=0,
        drops_spans=0,
    ):
        self.event_processor = event_processor
        self._update_resulting_calls(
            "event_processor",
            drops_transactions,
            drops_spans,
        )

        return self

    def run(self, sentry_init, capture_record_lost_event_calls):
        """Runs the test case with the configured parameters."""
        sentry_init(before_send_transaction=self.before_send)
        record_lost_event_calls = capture_record_lost_event_calls()

        with sentry_sdk.isolation_scope() as scope:
            scope.add_event_processor(self.event_processor)
            event = self.mock_transaction_event(self.span_count)
            sentry_sdk.get_client().capture_event(event, scope=scope)

        # We use counters to ensure that the calls are made the expected number of times, disregarding order.
        assert Counter(record_lost_event_calls) == self.expected_record_lost_event_calls


@pytest.mark.parametrize(
    "test_config",
    (
        TestSpanClientReports(span_count=10),  # No spans dropped
        TestSpanClientReports(span_count=0).with_before_send(
            lambda e, _: None,
            drops_transactions=1,
            drops_spans=1,
        ),
        TestSpanClientReports(span_count=10).with_before_send(
            lambda e, _: None,
            drops_transactions=1,
            drops_spans=11,
        ),
        TestSpanClientReports(span_count=10).with_before_send(
            TestSpanClientReports.span_dropper(3),
            drops_spans=3,
        ),
        TestSpanClientReports(span_count=10).with_before_send(
            TestSpanClientReports.span_dropper(10),
            drops_spans=10,
        ),
        TestSpanClientReports(span_count=10).with_event_processor(
            lambda e, _: None,
            drops_transactions=1,
            drops_spans=11,
        ),
        TestSpanClientReports(span_count=10).with_event_processor(
            TestSpanClientReports.span_dropper(3),
            drops_spans=3,
        ),
        TestSpanClientReports(span_count=10).with_event_processor(
            TestSpanClientReports.span_dropper(10),
            drops_spans=10,
        ),
        TestSpanClientReports(span_count=10)
        .with_event_processor(
            TestSpanClientReports.span_dropper(3),
            drops_spans=3,
        )
        .with_before_send(
            TestSpanClientReports.span_dropper(5),
            drops_spans=5,
        ),
        TestSpanClientReports(10)
        .with_event_processor(
            TestSpanClientReports.span_dropper(3),
            drops_spans=3,
        )
        .with_before_send(
            lambda e, _: None,
            drops_transactions=1,
            drops_spans=8,  # 3 of the 11 (incl. transaction) spans already dropped
        ),
    ),
)
def test_dropped_transaction(sentry_init, capture_record_lost_event_calls, test_config):
    test_config.run(sentry_init, capture_record_lost_event_calls)


@pytest.mark.parametrize("enable_tracing", [True, False])
def test_enable_tracing_deprecated(sentry_init, enable_tracing):
    with pytest.warns(DeprecationWarning):
        sentry_init(enable_tracing=enable_tracing)


def test_ignore_spans_warns_without_streaming(sentry_init):
    with pytest.warns(UserWarning, match=r"`ignore_spans` parameter only works"):
        sentry_init(ignore_spans=["/health"], trace_lifecycle="static")


@pytest.mark.parametrize(
    "options",
    [
        {"ignore_spans": ["/health"], "trace_lifecycle": "stream"},
        {"ignore_spans": ["/health"], "_experiments": {"trace_lifecycle": "stream"}},
        {},
    ],
)
def test_ignore_spans_does_not_warn(sentry_init, options):
    with warnings.catch_warnings(record=True) as caught:
        warnings.simplefilter("always")
        sentry_init(**options)

    ignore_spans_warnings = [w for w in caught if "ignore_spans" in str(w.message)]
    assert ignore_spans_warnings == []


def make_options_transport_cls():
    """Make an options transport class that captures the options passed to it."""
    # We need a unique class for each test so that the options are not
    # shared between tests.

    class OptionsTransport(Transport):
        """Transport that captures the options passed to it."""

        def __init__(self, options):
            super().__init__(options)
            type(self).options = options

        def capture_envelope(self, _):
            pass

    return OptionsTransport


@contextlib.contextmanager
def clear_env_var(name):
    """Helper to clear the a given environment variable,
    and restore it to its original value on exit."""
    old_value = os.environ.pop(name, None)

    try:
        yield
    finally:
        if old_value is not None:
            os.environ[name] = old_value
        elif name in os.environ:
            del os.environ[name]


@pytest.mark.parametrize(
    ("env_value", "arg_value", "expected_value"),
    [
        (None, None, False),  # default
        ("0", None, False),  # env var false
        ("1", None, True),  # env var true
        (None, False, False),  # arg false
        (None, True, True),  # arg true
        # Argument overrides environment variable
        ("0", True, True),  # env false, arg true
        ("1", False, False),  # env true, arg false
    ],
)
def test_keep_alive(env_value, arg_value, expected_value):
    transport_cls = make_options_transport_cls()
    keep_alive_kwarg = {} if arg_value is None else {"keep_alive": arg_value}

    with clear_env_var("SENTRY_KEEP_ALIVE"):
        if env_value is not None:
            os.environ["SENTRY_KEEP_ALIVE"] = env_value

        sentry_sdk.init(
            dsn="http://foo@sentry.io/123",
            transport=transport_cls,
            **keep_alive_kwarg,
        )

    assert transport_cls.options["keep_alive"] is expected_value


@pytest.mark.parametrize("testcase", PROXY_TESTCASES)
@skip_under_gevent
@pytest.mark.asyncio
@pytest.mark.skipif(not PY38, reason="Async transport requires Python 3.8+")
async def test_async_proxy(monkeypatch, testcase):
    # These are just the same tests as the sync ones, but they need to be run in an event loop
    # and respect the shutdown behavior of the async transport
    if testcase["env_http_proxy"] is not None:
        monkeypatch.setenv("HTTP_PROXY", testcase["env_http_proxy"])
    if testcase["env_https_proxy"] is not None:
        monkeypatch.setenv("HTTPS_PROXY", testcase["env_https_proxy"])
    if testcase.get("env_no_proxy") is not None:
        monkeypatch.setenv("NO_PROXY", testcase["env_no_proxy"])

    kwargs = {
        "_experiments": {"transport_async": True},
        "integrations": [AsyncioIntegration()],
    }

    if testcase["arg_http_proxy"] is not None:
        kwargs["http_proxy"] = testcase["arg_http_proxy"]
    if testcase["arg_https_proxy"] is not None:
        kwargs["https_proxy"] = testcase["arg_https_proxy"]
    if testcase.get("arg_proxy_headers") is not None:
        kwargs["proxy_headers"] = testcase["arg_proxy_headers"]

    client = Client(testcase["dsn"], **kwargs)
    assert isinstance(client.transport, AsyncHttpTransport)

    proxy = getattr(
        client.transport._pool,
        "proxy",
        getattr(client.transport._pool, "_proxy_url", None),
    )
    if testcase["expected_proxy_scheme"] is None:
        assert proxy is None
    else:
        scheme = (
            proxy.scheme.decode("ascii")
            if isinstance(proxy.scheme, bytes)
            else proxy.scheme
        )
        assert scheme == testcase["expected_proxy_scheme"]

        if testcase.get("arg_proxy_headers") is not None:
            proxy_headers = dict(
                (k.decode("ascii"), v.decode("ascii"))
                for k, v in client.transport._pool._proxy_headers
            )
            assert proxy_headers == testcase["arg_proxy_headers"]

    await client.close_async()


@skip_under_gevent
@pytest.mark.asyncio
@pytest.mark.skipif(not PY38, reason="Async client methods require Python 3.8+")
async def test_close_with_async_transport_warns():
    """Test close() with AsyncHttpTransport emits a warning."""
    import warnings as _warnings

    client = Client(
        "https://foo@sentry.io/123",
        _experiments={"transport_async": True},
        integrations=[AsyncioIntegration()],
    )
    assert isinstance(client.transport, AsyncHttpTransport)

    with _warnings.catch_warnings(record=True) as w:
        _warnings.simplefilter("always")
        client.close()
        assert any("close_async()" in str(warning.message) for warning in w)


@skip_under_gevent
@pytest.mark.asyncio
@pytest.mark.skipif(not PY38, reason="Async client methods require Python 3.8+")
async def test_close_async_with_async_transport():
    """Test close_async() properly closes async transport."""
    client = Client(
        "https://foo@sentry.io/123",
        _experiments={"transport_async": True},
        integrations=[AsyncioIntegration()],
    )
    assert isinstance(client.transport, AsyncHttpTransport)

    await client.close_async(timeout=1.0)
    assert client.transport is None


@skip_under_gevent
@pytest.mark.asyncio
@pytest.mark.skipif(not PY38, reason="Async client methods require Python 3.8+")
async def test_close_async_with_sync_transport():
    """Test close_async() aborts with non-async transport."""
    client = Client("https://foo@sentry.io/123")
    assert not isinstance(client.transport, AsyncHttpTransport)

    with mock.patch("sentry_sdk.client.logger") as mock_logger:
        await client.close_async()
        mock_logger.debug.assert_any_call(
            "close_async() used with non-async transport, aborting. Please use close() instead."
        )
    # Transport should NOT have been set to None
    assert client.transport is not None
    client.close()


@skip_under_gevent
@pytest.mark.asyncio
@pytest.mark.skipif(not PY38, reason="Async client methods require Python 3.8+")
async def test_close_async_no_transport():
    """Test close_async() does nothing when transport is None."""
    client = Client("https://foo@sentry.io/123")
    client.transport = None
    # Should not raise
    await client.close_async()


@skip_under_gevent
@pytest.mark.asyncio
@pytest.mark.skipif(not PY38, reason="Async client methods require Python 3.8+")
async def test_flush_with_async_transport_warns():
    """Test flush() with AsyncHttpTransport emits a warning and returns."""
    import warnings as _warnings

    client = Client(
        "https://foo@sentry.io/123",
        _experiments={"transport_async": True},
        integrations=[AsyncioIntegration()],
    )
    assert isinstance(client.transport, AsyncHttpTransport)

    with _warnings.catch_warnings(record=True) as w:
        _warnings.simplefilter("always")
        client.flush(timeout=1.0)
        assert any("flush_async()" in str(warning.message) for warning in w)
    await client.close_async()


@skip_under_gevent
@pytest.mark.asyncio
@pytest.mark.skipif(not PY38, reason="Async client methods require Python 3.8+")
async def test_flush_async_with_async_transport():
    """Test flush_async() works with async transport."""
    client = Client(
        "https://foo@sentry.io/123",
        _experiments={"transport_async": True},
        integrations=[AsyncioIntegration()],
    )
    assert isinstance(client.transport, AsyncHttpTransport)

    # Should not raise
    await client.flush_async(timeout=1.0)
    await client.close_async()


@skip_under_gevent
@pytest.mark.asyncio
@pytest.mark.skipif(not PY38, reason="Async client methods require Python 3.8+")
async def test_flush_async_uses_shutdown_timeout_default():
    """Test flush_async() uses shutdown_timeout when no timeout provided."""
    client = Client(
        "https://foo@sentry.io/123",
        _experiments={"transport_async": True},
        integrations=[AsyncioIntegration()],
        shutdown_timeout=5.0,
    )
    assert isinstance(client.transport, AsyncHttpTransport)

    with mock.patch.object(client.transport, "flush", return_value=None) as mock_flush:
        await client.flush_async()
        mock_flush.assert_called_once_with(timeout=5.0, callback=None)
    await client.close_async()


@skip_under_gevent
@pytest.mark.asyncio
@pytest.mark.skipif(not PY38, reason="Async client methods require Python 3.8+")
async def test_flush_async_with_sync_transport():
    """Test flush_async() aborts with non-async transport."""
    client = Client("https://foo@sentry.io/123")
    assert not isinstance(client.transport, AsyncHttpTransport)

    with mock.patch("sentry_sdk.client.logger") as mock_logger:
        await client.flush_async()
        mock_logger.debug.assert_any_call(
            "flush_async() used with non-async transport, aborting. Please use flush() instead."
        )
    client.close()


@skip_under_gevent
@pytest.mark.asyncio
@pytest.mark.skipif(not PY38, reason="Async client methods require Python 3.8+")
async def test_flush_async_no_transport():
    """Test flush_async() does nothing when transport is None."""
    client = Client("https://foo@sentry.io/123")
    client.transport = None
    # Should not raise
    await client.flush_async()


@skip_under_gevent
@pytest.mark.asyncio
@pytest.mark.skipif(not PY38, reason="Async client methods require Python 3.8+")
async def test_client_async_context_manager():
    """Test Client works as async context manager."""
    async with Client(
        "https://foo@sentry.io/123",
        _experiments={"transport_async": True},
        integrations=[AsyncioIntegration()],
    ) as client:
        assert isinstance(client.transport, AsyncHttpTransport)
        assert client.is_active()
    # After __aexit__, transport should be None
    assert client.transport is None


@skip_under_gevent
@pytest.mark.asyncio
@pytest.mark.skipif(not PY38, reason="Async client methods require Python 3.8+")
async def test_client_async_context_manager_with_sync_transport():
    """Test Client async context manager with sync transport is safe."""
    async with Client("https://foo@sentry.io/123") as client:
        assert client.is_active()
    # close_async with sync transport is a no-op for transport cleanup
    # Transport won't be None because close_async aborts for non-async transport
    assert client.transport is not None
    client.close()


@pytest.mark.parametrize("testcase", SOCKS_PROXY_TESTCASES)
@skip_under_gevent
@pytest.mark.asyncio
@pytest.mark.skipif(not PY38, reason="Async transport requires Python 3.8+")
async def test_async_socks_proxy(testcase):
    # These are just the same tests as the sync ones, but they need to be run in an event loop
    # and respect the shutdown behavior of the async transport

    kwargs = {
        "_experiments": {"transport_async": True},
        "integrations": [AsyncioIntegration()],
    }

    if testcase["arg_http_proxy"] is not None:
        kwargs["http_proxy"] = testcase["arg_http_proxy"]
    if testcase["arg_https_proxy"] is not None:
        kwargs["https_proxy"] = testcase["arg_https_proxy"]

    client = Client(testcase["dsn"], **kwargs)
    assert isinstance(client.transport, AsyncHttpTransport)

    assert ("socks" in str(type(client.transport._pool)).lower()) == testcase[
        "should_be_socks_proxy"
    ], (
        f"Expected {kwargs} to result in SOCKS == {testcase['should_be_socks_proxy']}"
        f"but got {str(type(client.transport._pool))}"
    )

    await client.close_async()



from dataclasses import asdict, dataclass
from typing import List, Optional

import pytest

from sentry_sdk.tracing_utils import (
    Baggage,
    _should_be_included,
    _should_continue_trace,
)
from tests.conftest import TestTransportWithOptions


def id_function(val: object) -> str:
    if isinstance(val, ShouldBeIncludedTestCase):
        return val.id


@dataclass(frozen=True)
class ShouldBeIncludedTestCase:
    id: str
    is_sentry_sdk_frame: bool
    namespace: Optional[str] = None
    in_app_include: Optional[List[str]] = None
    in_app_exclude: Optional[List[str]] = None
    abs_path: Optional[str] = None
    project_root: Optional[str] = None


@pytest.mark.parametrize(
    "test_case, expected",
    [
        (
            ShouldBeIncludedTestCase(
                id="Frame from Sentry SDK",
                is_sentry_sdk_frame=True,
            ),
            False,
        ),
        (
            ShouldBeIncludedTestCase(
                id="Frame from Django installed in virtualenv inside project root",
                is_sentry_sdk_frame=False,
                abs_path="/home/username/some_project/.venv/lib/python3.12/site-packages/django/db/models/sql/compiler",
                project_root="/home/username/some_project",
                namespace="django.db.models.sql.compiler",
                in_app_include=["django"],
            ),
            True,
        ),
        (
            ShouldBeIncludedTestCase(
                id="Frame from project",
                is_sentry_sdk_frame=False,
                abs_path="/home/username/some_project/some_project/__init__.py",
                project_root="/home/username/some_project",
                namespace="some_project",
            ),
            True,
        ),
        (
            ShouldBeIncludedTestCase(
                id="Frame from project module in `in_app_exclude`",
                is_sentry_sdk_frame=False,
                abs_path="/home/username/some_project/some_project/exclude_me/some_module.py",
                project_root="/home/username/some_project",
                namespace="some_project.exclude_me.some_module",
                in_app_exclude=["some_project.exclude_me"],
            ),
            False,
        ),
        (
            ShouldBeIncludedTestCase(
                id="Frame from system-wide installed Django",
                is_sentry_sdk_frame=False,
                abs_path="/usr/lib/python3.12/site-packages/django/db/models/sql/compiler",
                project_root="/home/username/some_project",
                namespace="django.db.models.sql.compiler",
            ),
            False,
        ),
        (
            ShouldBeIncludedTestCase(
                id="Frame from system-wide installed Django with `django` in `in_app_include`",
                is_sentry_sdk_frame=False,
                abs_path="/usr/lib/python3.12/site-packages/django/db/models/sql/compiler",
                project_root="/home/username/some_project",
                namespace="django.db.models.sql.compiler",
                in_app_include=["django"],
            ),
            True,
        ),
    ],
    ids=id_function,
)
def test_should_be_included(
    test_case: "ShouldBeIncludedTestCase", expected: bool
) -> None:
    """Checking logic, see: https://github.com/getsentry/sentry-python/issues/3312"""
    kwargs = asdict(test_case)
    kwargs.pop("id")
    assert _should_be_included(**kwargs) == expected


@pytest.mark.parametrize(
    ("header", "expected"),
    (
        ("", ""),
        ("foo=bar", "foo=bar"),
        (" foo=bar, baz =  qux ", " foo=bar, baz =  qux "),
        ("sentry-trace_id=123", ""),
        ("  sentry-trace_id = 123  ", ""),
        ("sentry-trace_id=123,sentry-public_key=456", ""),
        ("foo=bar,sentry-trace_id=123", "foo=bar"),
        ("foo=bar,sentry-trace_id=123,baz=qux", "foo=bar,baz=qux"),
        (
            "foo=bar,sentry-trace_id=123,baz=qux,sentry-public_key=456",
            "foo=bar,baz=qux",
        ),
    ),
)
def test_strip_sentry_baggage(header, expected):
    assert Baggage.strip_sentry_baggage(header) == expected


@pytest.mark.parametrize(
    ("baggage", "expected_repr"),
    (
        (Baggage(sentry_items={}), '<Baggage "", mutable=True>'),
        (Baggage(sentry_items={}, mutable=False), '<Baggage "", mutable=False>'),
        (
            Baggage(sentry_items={"foo": "bar"}),
            '<Baggage "sentry-foo=bar,", mutable=True>',
        ),
        (
            Baggage(sentry_items={"foo": "bar"}, mutable=False),
            '<Baggage "sentry-foo=bar,", mutable=False>',
        ),
        (
            Baggage(sentry_items={"foo": "bar"}, third_party_items="asdf=1234,"),
            '<Baggage "sentry-foo=bar,asdf=1234,", mutable=True>',
        ),
        (
            Baggage(
                sentry_items={"foo": "bar"},
                third_party_items="asdf=1234,",
                mutable=False,
            ),
            '<Baggage "sentry-foo=bar,asdf=1234,", mutable=False>',
        ),
    ),
)
def test_baggage_repr(baggage, expected_repr):
    assert repr(baggage) == expected_repr


@pytest.mark.parametrize(
    (
        "baggage_header",
        "dsn",
        "explicit_org_id",
        "strict_trace_continuation",
        "should_continue_trace",
    ),
    (
        # continue cases when strict_trace_continuation=False
        (
            "sentry-trace_id=771a43a4192642f0b136d5159a501700, sentry-org_id=1234",
            "https://mysecret@o1234.ingest.sentry.io/12312012",
            None,
            False,
            True,
        ),
        (
            "sentry-trace_id=771a43a4192642f0b136d5159a501700",
            "https://mysecret@o1234.ingest.sentry.io/12312012",
            None,
            False,
            True,
        ),
        (
            "sentry-trace_id=771a43a4192642f0b136d5159a501700, sentry-org_id=1234",
            None,
            None,
            False,
            True,
        ),
        (
            "sentry-trace_id=771a43a4192642f0b136d5159a501700, sentry-org_id=1234",
            None,
            "1234",
            False,
            True,
        ),
        (
            "sentry-trace_id=771a43a4192642f0b136d5159a501700, sentry-org_id=1234",
            "https://mysecret@not_org_id.ingest.sentry.io/12312012",
            None,
            False,
            True,
        ),
        # start new cases when strict_trace_continuation=False
        (
            "sentry-trace_id=771a43a4192642f0b136d5159a501700, sentry-org_id=1234",
            "https://mysecret@o9999.ingest.sentry.io/12312012",
            None,
            False,
            False,
        ),
        (
            "sentry-trace_id=771a43a4192642f0b136d5159a501700, sentry-org_id=1234",
            "https://mysecret@o1234.ingest.sentry.io/12312012",
            "9999",
            False,
            False,
        ),
        # continue cases when strict_trace_continuation=True
        (
            "sentry-trace_id=771a43a4192642f0b136d5159a501700, sentry-org_id=1234",
            "https://mysecret@o1234.ingest.sentry.io/12312012",
            None,
            True,
            True,
        ),
        ("sentry-trace_id=771a43a4192642f0b136d5159a501700", None, None, True, True),
        # start new cases when strict_trace_continuation=True
        (
            "sentry-trace_id=771a43a4192642f0b136d5159a501700",
            "https://mysecret@o1234.ingest.sentry.io/12312012",
            None,
            True,
            False,
        ),
        (
            "sentry-trace_id=771a43a4192642f0b136d5159a501700, sentry-org_id=1234",
            None,
            None,
            True,
            False,
        ),
        (
            "sentry-trace_id=771a43a4192642f0b136d5159a501700, sentry-org_id=1234",
            "https://mysecret@not_org_id.ingest.sentry.io/12312012",
            None,
            True,
            False,
        ),
        (
            "sentry-trace_id=771a43a4192642f0b136d5159a501700, sentry-org_id=1234",
            "https://mysecret@o9999.ingest.sentry.io/12312012",
            None,
            True,
            False,
        ),
        (
            "sentry-trace_id=771a43a4192642f0b136d5159a501700, sentry-org_id=1234",
            "https://mysecret@o1234.ingest.sentry.io/12312012",
            "9999",
            True,
            False,
        ),
    ),
)
def test_should_continue_trace(
    sentry_init,
    baggage_header,
    dsn,
    explicit_org_id,
    strict_trace_continuation,
    should_continue_trace,
):
    sentry_init(
        dsn=dsn,
        org_id=explicit_org_id,
        strict_trace_continuation=strict_trace_continuation,
        traces_sample_rate=1.0,
        transport=TestTransportWithOptions,
    )

    baggage = Baggage.from_incoming_header(baggage_header) if baggage_header else None
    assert _should_continue_trace(baggage) == should_continue_trace


def test_baggage_from_incoming_header_value_with_equals_sign():
    # Baggage values that legitimately contain '=' (e.g. base64) must not be dropped
    header = "sentry-release=v1.0==1,sentry-trace_id=abc123"
    baggage = Baggage.from_incoming_header(header)
    assert baggage.sentry_items == {"release": "v1.0==1", "trace_id": "abc123"}



import ast
import importlib
import pathlib
import pkgutil
import sys
import types

import pytest

from sentry_sdk import integrations


def pytest_generate_tests(metafunc):
    """
    All submodules of sentry_sdk.integrations are picked up, so modules
    without a subclass of sentry_sdk.integrations.Integration are also tested
    for poorly gated imports.

    This approach was chosen to keep the implementation simple.
    """
    if "integration_submodule_name" in metafunc.fixturenames:
        submodule_names = {
            submodule_name
            for _, submodule_name, _ in pkgutil.walk_packages(integrations.__path__)
        }

        metafunc.parametrize(
            "integration_submodule_name",
            submodule_names,
        )


def find_unrecognized_dependencies(tree):
    """
    Finds unrecognized imports in the AST for a Python module. In an empty
    environment the set of non-standard library modules is returned.
    """
    unrecognized_dependencies = set()
    package_name = lambda name: name.split(".")[0]

    for node in ast.walk(tree):
        if isinstance(node, ast.Import):
            for alias in node.names:
                root = package_name(alias.name)

                try:
                    if not importlib.util.find_spec(root):
                        unrecognized_dependencies.add(root)
                except ValueError:
                    continue

        elif isinstance(node, ast.ImportFrom):
            # if node.level is not 0 the import is relative
            if node.level > 0 or node.module is None:
                continue

            root = package_name(node.module)

            try:
                if not importlib.util.find_spec(root):
                    unrecognized_dependencies.add(root)
            except ValueError:
                continue

    return unrecognized_dependencies


@pytest.mark.skipif(
    sys.version_info < (3, 7), reason="asyncpg imports __future__.annotations"
)
def test_shadowed_modules_when_importing_integrations(
    sentry_init, integration_submodule_name
):
    """
    Check that importing integrations for third-party module raises an
    DidNotEnable exception when the associated module is shadowed by an empty
    module.

    An integration is determined to be for a third-party module if it cannot
    be imported in the environment in which the tests run.
    """
    module_path = f"sentry_sdk.integrations.{integration_submodule_name}"
    try:
        # If importing the integration succeeds in the current environment, assume
        # that the integration has no non-standard imports.
        importlib.import_module(module_path)
        return
    except integrations.DidNotEnable:
        spec = importlib.util.find_spec(module_path)
        source = pathlib.Path(spec.origin).read_text(encoding="utf-8")
        tree = ast.parse(source, filename=spec.origin)
        integration_dependencies = find_unrecognized_dependencies(tree)

        # For each non-standard import, create an empty shadow module to
        # emulate an empty "agents.py" or analogous local module that
        # shadows the package.
        for dependency in integration_dependencies:
            sys.modules[dependency] = types.ModuleType(dependency)

        # Importing the integration must raise DidNotEnable, since the
        # SDK catches the exception type when attempting to activate
        # auto-enabling integrations.
        with pytest.raises(integrations.DidNotEnable):
            importlib.import_module(module_path)

        for dependency in integration_dependencies:
            del sys.modules[dependency]



import threading
import time
from collections import defaultdict
from unittest import mock

import pytest

import sentry_sdk
from sentry_sdk.consts import VERSION
from sentry_sdk.profiler.continuous_profiler import (
    get_profiler_id,
    is_profile_session_sampled,
    setup_continuous_profiler,
    start_profile_session,
    start_profiler,
    stop_profile_session,
    stop_profiler,
)
from tests.conftest import ApproxDict

try:
    import gevent
except ImportError:
    gevent = None


requires_gevent = pytest.mark.skipif(gevent is None, reason="gevent not enabled")


def get_client_options(use_top_level_profiler_mode):
    def client_options(
        mode=None, auto_start=None, profile_session_sample_rate=1.0, lifecycle="manual"
    ):
        if use_top_level_profiler_mode:
            return {
                "profile_lifecycle": lifecycle,
                "profiler_mode": mode,
                "profile_session_sample_rate": profile_session_sample_rate,
                "_experiments": {
                    "continuous_profiling_auto_start": auto_start,
                },
            }
        return {
            "profile_lifecycle": lifecycle,
            "profile_session_sample_rate": profile_session_sample_rate,
            "_experiments": {
                "continuous_profiling_auto_start": auto_start,
                "continuous_profiling_mode": mode,
            },
        }

    return client_options


mock_sdk_info = {
    "name": "sentry.python",
    "version": VERSION,
    "packages": [{"name": "pypi:sentry-sdk", "version": VERSION}],
}


@pytest.mark.parametrize("mode", [pytest.param("foo")])
@pytest.mark.parametrize(
    "make_options",
    [
        pytest.param(get_client_options(True), id="non-experiment"),
        pytest.param(get_client_options(False), id="experiment"),
    ],
)
def test_continuous_profiler_invalid_mode(mode, make_options, teardown_profiling):
    with pytest.raises(ValueError):
        setup_continuous_profiler(
            make_options(mode=mode),
            mock_sdk_info,
            lambda envelope: None,
        )


@pytest.mark.parametrize(
    "mode",
    [
        pytest.param("thread"),
        pytest.param("gevent", marks=requires_gevent),
    ],
)
@pytest.mark.parametrize(
    "make_options",
    [
        pytest.param(get_client_options(True), id="non-experiment"),
        pytest.param(get_client_options(False), id="experiment"),
    ],
)
def test_continuous_profiler_valid_mode(mode, make_options, teardown_profiling):
    options = make_options(mode=mode)
    setup_continuous_profiler(
        options,
        mock_sdk_info,
        lambda envelope: None,
    )


@pytest.mark.parametrize(
    "mode",
    [
        pytest.param("thread"),
        pytest.param("gevent", marks=requires_gevent),
    ],
)
@pytest.mark.parametrize(
    "make_options",
    [
        pytest.param(get_client_options(True), id="non-experiment"),
        pytest.param(get_client_options(False), id="experiment"),
    ],
)
def test_continuous_profiler_setup_twice(mode, make_options, teardown_profiling):
    assert not is_profile_session_sampled()

    # setting up the first time should return True to indicate success
    options = make_options(mode=mode, profile_session_sample_rate=1.0)
    assert setup_continuous_profiler(
        options,
        mock_sdk_info,
        lambda envelope: None,
    )
    assert is_profile_session_sampled()

    # setting up the second time should return True to indicate re-init
    options = make_options(mode=mode, profile_session_sample_rate=0.0)
    assert setup_continuous_profiler(
        options,
        mock_sdk_info,
        lambda envelope: None,
    )
    assert not is_profile_session_sampled()


def assert_single_transaction_with_profile_chunks(
    envelopes, thread, max_chunks=None, transactions=1
):
    items = defaultdict(list)
    for envelope in envelopes:
        for item in envelope.items:
            items[item.type].append(item)

    assert len(items["transaction"]) == transactions
    assert len(items["profile_chunk"]) > 0
    if max_chunks is not None:
        assert len(items["profile_chunk"]) <= max_chunks

    for chunk_item in items["profile_chunk"]:
        chunk = chunk_item.payload.json
        headers = chunk_item.headers
        assert chunk["platform"] == headers["platform"]

    transaction = items["transaction"][0].payload.json

    trace_context = transaction["contexts"]["trace"]

    assert trace_context == ApproxDict(
        {
            "data": ApproxDict(
                {
                    "thread.id": str(thread.ident),
                    "thread.name": thread.name,
                }
            ),
        }
    )

    profile_context = transaction["contexts"]["profile"]
    profiler_id = profile_context["profiler_id"]

    assert profile_context == ApproxDict({"profiler_id": profiler_id})

    spans = transaction["spans"]
    assert len(spans) > 0
    for span in spans:
        assert span["data"] == ApproxDict(
            {
                "profiler_id": profiler_id,
                "thread.id": str(thread.ident),
                "thread.name": thread.name,
            }
        )

    for profile_chunk_item in items["profile_chunk"]:
        profile_chunk = profile_chunk_item.payload.json
        del profile_chunk["profile"]  # make the diff easier to read
        assert profile_chunk == ApproxDict(
            {
                "client_sdk": {
                    "name": mock.ANY,
                    "version": VERSION,
                },
                "platform": "python",
                "profiler_id": profiler_id,
                "version": "2",
            }
        )


def assert_single_transaction_without_profile_chunks(envelopes):
    items = defaultdict(list)
    for envelope in envelopes:
        for item in envelope.items:
            items[item.type].append(item)

    assert len(items["transaction"]) == 1
    assert len(items["profile_chunk"]) == 0

    transaction = items["transaction"][0].payload.json
    assert "profile" not in transaction["contexts"]


@pytest.mark.forked
@pytest.mark.parametrize(
    "mode",
    [
        pytest.param("thread"),
        pytest.param("gevent", marks=requires_gevent),
    ],
)
@pytest.mark.parametrize(
    ["start_profiler_func", "stop_profiler_func"],
    [
        pytest.param(
            start_profile_session,
            stop_profile_session,
            id="start_profile_session/stop_profile_session (deprecated)",
        ),
        pytest.param(
            start_profiler,
            stop_profiler,
            id="start_profiler/stop_profiler",
        ),
    ],
)
@pytest.mark.parametrize(
    "make_options",
    [
        pytest.param(get_client_options(True), id="non-experiment"),
        pytest.param(get_client_options(False), id="experiment"),
    ],
)
def test_continuous_profiler_auto_start_and_manual_stop(
    sentry_init,
    capture_envelopes,
    mode,
    start_profiler_func,
    stop_profiler_func,
    make_options,
    teardown_profiling,
):
    options = make_options(mode=mode, auto_start=True)
    sentry_init(
        traces_sample_rate=1.0,
        **options,
    )

    envelopes = capture_envelopes()

    thread = threading.current_thread()

    with sentry_sdk.start_transaction(name="profiling"):
        with sentry_sdk.start_span(op="op"):
            pass

    for _ in range(3):
        stop_profiler_func()

        assert_single_transaction_with_profile_chunks(envelopes, thread)

        envelopes.clear()

        with sentry_sdk.start_transaction(name="profiling"):
            with sentry_sdk.start_span(op="op"):
                time.sleep(0.05)

        assert_single_transaction_without_profile_chunks(envelopes)

        start_profiler_func()

        envelopes.clear()

        with sentry_sdk.start_transaction(name="profiling"):
            with sentry_sdk.start_span(op="op"):
                pass

    stop_profiler_func()

    assert_single_transaction_with_profile_chunks(envelopes, thread)


@pytest.mark.parametrize(
    "mode",
    [
        pytest.param("thread"),
        pytest.param("gevent", marks=requires_gevent),
    ],
)
@pytest.mark.parametrize(
    ["start_profiler_func", "stop_profiler_func"],
    [
        pytest.param(
            start_profile_session,
            stop_profile_session,
            id="start_profile_session/stop_profile_session  (deprecated)",
        ),
        pytest.param(
            start_profiler,
            stop_profiler,
            id="start_profiler/stop_profiler",
        ),
    ],
)
@pytest.mark.parametrize(
    "make_options",
    [
        pytest.param(get_client_options(True), id="non-experiment"),
        pytest.param(get_client_options(False), id="experiment"),
    ],
)
@mock.patch("sentry_sdk.profiler.continuous_profiler.PROFILE_BUFFER_SECONDS", 0.01)
def test_continuous_profiler_manual_start_and_stop_sampled(
    sentry_init,
    capture_envelopes,
    mode,
    start_profiler_func,
    stop_profiler_func,
    make_options,
    teardown_profiling,
):
    options = make_options(
        mode=mode, profile_session_sample_rate=1.0, lifecycle="manual"
    )
    sentry_init(
        traces_sample_rate=1.0,
        **options,
    )

    envelopes = capture_envelopes()

    thread = threading.current_thread()

    for _ in range(3):
        start_profiler_func()

        envelopes.clear()

        with sentry_sdk.start_transaction(name="profiling"):
            assert get_profiler_id() is not None, "profiler should be running"
            with sentry_sdk.start_span(op="op"):
                time.sleep(0.1)
            assert get_profiler_id() is not None, "profiler should be running"

        assert get_profiler_id() is not None, "profiler should be running"

        stop_profiler_func()

        assert_single_transaction_with_profile_chunks(envelopes, thread)

        # the profiler stops immediately in manual mode
        assert get_profiler_id() is None, "profiler should not be running"

        envelopes.clear()

        with sentry_sdk.start_transaction(name="profiling"):
            assert get_profiler_id() is None, "profiler should not be running"
            with sentry_sdk.start_span(op="op"):
                time.sleep(0.1)
            assert get_profiler_id() is None, "profiler should not be running"

        assert_single_transaction_without_profile_chunks(envelopes)


@pytest.mark.parametrize(
    "mode",
    [
        pytest.param("thread"),
        pytest.param("gevent", marks=requires_gevent),
    ],
)
@pytest.mark.parametrize(
    ["start_profiler_func", "stop_profiler_func"],
    [
        pytest.param(
            start_profile_session,
            stop_profile_session,
            id="start_profile_session/stop_profile_session (deprecated)",
        ),
        pytest.param(
            start_profiler,
            stop_profiler,
            id="start_profiler/stop_profiler",
        ),
    ],
)
@pytest.mark.parametrize(
    "make_options",
    [
        pytest.param(get_client_options(True), id="non-experiment"),
        pytest.param(get_client_options(False), id="experiment"),
    ],
)
def test_continuous_profiler_manual_start_and_stop_unsampled(
    sentry_init,
    capture_envelopes,
    mode,
    start_profiler_func,
    stop_profiler_func,
    make_options,
    teardown_profiling,
):
    options = make_options(
        mode=mode, profile_session_sample_rate=0.0, lifecycle="manual"
    )
    sentry_init(
        traces_sample_rate=1.0,
        **options,
    )

    envelopes = capture_envelopes()

    start_profiler_func()

    with sentry_sdk.start_transaction(name="profiling"):
        with sentry_sdk.start_span(op="op"):
            pass

    stop_profiler_func()

    assert_single_transaction_without_profile_chunks(envelopes)


@pytest.mark.parametrize(
    "mode",
    [
        pytest.param("thread"),
        pytest.param("gevent", marks=requires_gevent),
    ],
)
@pytest.mark.parametrize(
    "make_options",
    [
        pytest.param(get_client_options(True), id="non-experiment"),
        pytest.param(get_client_options(False), id="experiment"),
    ],
)
@mock.patch("sentry_sdk.profiler.continuous_profiler.DEFAULT_SAMPLING_FREQUENCY", 21)
def test_continuous_profiler_auto_start_and_stop_sampled(
    sentry_init,
    capture_envelopes,
    mode,
    make_options,
    teardown_profiling,
):
    options = make_options(
        mode=mode, profile_session_sample_rate=1.0, lifecycle="trace"
    )
    sentry_init(
        traces_sample_rate=1.0,
        **options,
    )

    envelopes = capture_envelopes()

    thread = threading.current_thread()

    all_profiler_ids = set()

    for _ in range(3):
        envelopes.clear()

        profiler_ids = set()

        with sentry_sdk.start_transaction(name="profiling 1"):
            profiler_id = get_profiler_id()
            assert profiler_id is not None, "profiler should be running"
            profiler_ids.add(profiler_id)
            with sentry_sdk.start_span(op="op"):
                time.sleep(0.1)
            profiler_id = get_profiler_id()
            assert profiler_id is not None, "profiler should be running"
            profiler_ids.add(profiler_id)

        time.sleep(0.03)

        # the profiler takes a while to stop in auto mode so if we start
        # a transaction immediately, it'll be part of the same chunk
        profiler_id = get_profiler_id()
        assert profiler_id is not None, "profiler should be running"
        profiler_ids.add(profiler_id)

        with sentry_sdk.start_transaction(name="profiling 2"):
            profiler_id = get_profiler_id()
            assert profiler_id is not None, "profiler should be running"
            profiler_ids.add(profiler_id)
            with sentry_sdk.start_span(op="op"):
                time.sleep(0.1)
            profiler_id = get_profiler_id()
            assert profiler_id is not None, "profiler should be running"
            profiler_ids.add(profiler_id)

        # wait at least 1 cycle for the profiler to stop
        time.sleep(0.2)
        assert get_profiler_id() is None, "profiler should not be running"

        assert len(profiler_ids) == 1
        all_profiler_ids.add(profiler_ids.pop())

        assert_single_transaction_with_profile_chunks(
            envelopes, thread, max_chunks=1, transactions=2
        )

    assert len(all_profiler_ids) == 3


@pytest.mark.parametrize(
    "mode",
    [
        pytest.param("thread"),
        pytest.param("gevent", marks=requires_gevent),
    ],
)
@pytest.mark.parametrize(
    "make_options",
    [
        pytest.param(get_client_options(True), id="non-experiment"),
        pytest.param(get_client_options(False), id="experiment"),
    ],
)
@mock.patch("sentry_sdk.profiler.continuous_profiler.PROFILE_BUFFER_SECONDS", 0.01)
def test_continuous_profiler_auto_start_and_stop_unsampled(
    sentry_init,
    capture_envelopes,
    mode,
    make_options,
    teardown_profiling,
):
    options = make_options(
        mode=mode, profile_session_sample_rate=0.0, lifecycle="trace"
    )
    sentry_init(
        traces_sample_rate=1.0,
        **options,
    )

    envelopes = capture_envelopes()

    for _ in range(3):
        envelopes.clear()

        with sentry_sdk.start_transaction(name="profiling"):
            assert get_profiler_id() is None, "profiler should not be running"
            with sentry_sdk.start_span(op="op"):
                time.sleep(0.05)
            assert get_profiler_id() is None, "profiler should not be running"

        assert get_profiler_id() is None, "profiler should not be running"
        assert_single_transaction_without_profile_chunks(envelopes)


@pytest.mark.parametrize(
    ["mode", "class_name"],
    [
        pytest.param("thread", "ThreadContinuousScheduler"),
        pytest.param(
            "gevent",
            "GeventContinuousScheduler",
            marks=requires_gevent,
        ),
    ],
)
@pytest.mark.parametrize(
    ["start_profiler_func", "stop_profiler_func"],
    [
        pytest.param(
            start_profile_session,
            stop_profile_session,
            id="start_profile_session/stop_profile_session (deprecated)",
        ),
        pytest.param(
            start_profiler,
            stop_profiler,
            id="start_profiler/stop_profiler",
        ),
    ],
)
@pytest.mark.parametrize(
    "make_options",
    [
        pytest.param(get_client_options(True), id="non-experiment"),
        pytest.param(get_client_options(False), id="experiment"),
    ],
)
def test_continuous_profiler_manual_start_and_stop_noop_when_using_trace_lifecyle(
    sentry_init,
    mode,
    start_profiler_func,
    stop_profiler_func,
    class_name,
    make_options,
    teardown_profiling,
):
    options = make_options(
        mode=mode, profile_session_sample_rate=0.0, lifecycle="trace"
    )
    sentry_init(
        traces_sample_rate=1.0,
        **options,
    )

    with mock.patch(
        f"sentry_sdk.profiler.continuous_profiler.{class_name}.ensure_running"
    ) as mock_ensure_running:
        start_profiler_func()
        mock_ensure_running.assert_not_called()

    with mock.patch(
        f"sentry_sdk.profiler.continuous_profiler.{class_name}.teardown"
    ) as mock_teardown:
        stop_profiler_func()
        mock_teardown.assert_not_called()


def test_continuous_profiler_run_does_not_null_buffer(
    sentry_init,
    capture_envelopes,
    teardown_profiling,
):
    """
    Verifies that ContinuousScheduler.run() does not set self.buffer = None
    after exiting its sampling loop.

    Previously, run() would execute `self.buffer = None` after the while
    loop exited. During rapid stop/start cycles, this could race with
    ensure_running() which creates a new buffer: the old thread's cleanup
    would destroy the newly-created buffer, causing the new profiler thread
    to silently drop all samples (self.buffer is None in the sampler).

    The fix uses a local buffer reference for flushing and never sets
    self.buffer = None from run().
    """
    from sentry_sdk.profiler import continuous_profiler as cp

    options = get_client_options(True)(
        mode="thread", profile_session_sample_rate=1.0, lifecycle="manual"
    )
    sentry_init(traces_sample_rate=1.0, **options)
    envelopes = capture_envelopes()
    thread = threading.current_thread()

    # Start and verify profiler works
    start_profiler()
    envelopes.clear()
    with sentry_sdk.start_transaction(name="profiling"):
        with sentry_sdk.start_span(op="op"):
            pass

    # Get the scheduler and create a sentinel buffer.
    # We'll call run() directly to verify it doesn't null out self.buffer.
    scheduler = cp._scheduler
    assert scheduler is not None

    # Stop the profiler so the thread exits cleanly
    stop_profiler()

    assert_single_transaction_with_profile_chunks(envelopes, thread)

    # Now set up a fresh buffer and mark the scheduler as not running
    # (simulating the state right after ensure_running() created a new buffer
    # but the old thread hasn't done cleanup yet).
    scheduler.reset_buffer()
    buffer_before = scheduler.buffer
    assert buffer_before is not None

    # Simulate what happens when run() exits its while loop:
    # self.running is already False, so the while loop exits immediately.
    scheduler.running = False
    scheduler.run()

    # After the fix, run() should NOT have set self.buffer = None.
    # It should only flush using a local reference.
    assert scheduler.buffer is not None, (
        "run() must not set self.buffer = None; "
        "this would destroy buffers created by concurrent ensure_running() calls"
    )



import asyncio
from typing import Optional
from unittest import mock

import grpc
import pytest
import pytest_asyncio

import sentry_sdk
from sentry_sdk import start_span, start_transaction
from sentry_sdk.consts import OP
from sentry_sdk.integrations.grpc import GRPCIntegration
from sentry_sdk.tracing_utils import has_span_streaming_enabled
from tests.conftest import ApproxDict
from tests.integrations.grpc.grpc_test_service_pb2 import gRPCTestMessage
from tests.integrations.grpc.grpc_test_service_pb2_grpc import (
    add_gRPCTestServiceServicer_to_server,
    gRPCTestServiceServicer,
    gRPCTestServiceStub,
)


@pytest_asyncio.fixture(scope="function")
async def grpc_server_and_channel(sentry_init):
    """
    Creates an async gRPC server and a channel connected to it.
    Returns both for use in tests, and cleans up afterward.
    """
    channel: "Optional[grpc.aio.Channel]" = None
    server: "Optional[grpc.aio.Server]" = None

    async def inner(span_streaming: bool):
        nonlocal server, channel

        sentry_init(
            traces_sample_rate=1.0,
            integrations=[GRPCIntegration()],
            trace_lifecycle="stream" if span_streaming else "static",
        )

        # Create server
        server = grpc.aio.server()

        # Let gRPC choose a free port instead of hardcoding it
        port = server.add_insecure_port("[::]:0")

        # Add service implementation
        add_gRPCTestServiceServicer_to_server(TestService, server)

        # Start the server
        await asyncio.create_task(server.start())

        # Create channel connected to our server
        channel = grpc.aio.insecure_channel(f"localhost:{port}")  # noqa: E231

        return server, channel

    try:
        yield inner
    finally:
        if channel is not None:
            await channel.close()

        if server is not None:
            await server.stop(None)


@pytest.mark.asyncio
@pytest.mark.parametrize("span_streaming", [True, False])
async def test_noop_for_unimplemented_method(
    sentry_init,
    capture_events,
    capture_items,
    span_streaming,
):
    sentry_init(
        traces_sample_rate=1.0,
        integrations=[GRPCIntegration()],
        trace_lifecycle="stream" if span_streaming else "static",
    )

    # Create empty server with no services
    server = grpc.aio.server()
    port = server.add_insecure_port("[::]:0")  # Let gRPC choose a free port
    await asyncio.create_task(server.start())

    if span_streaming:
        items = capture_items("span")

        try:
            async with grpc.aio.insecure_channel(
                f"localhost:{port}"  # noqa: E231
            ) as channel:
                stub = gRPCTestServiceStub(channel)
                with pytest.raises(grpc.RpcError) as exc:
                    await stub.TestServe(gRPCTestMessage(text="test"))
                assert exc.value.details() == "Method not found!"
        finally:
            await server.stop(None)

        sentry_sdk.flush()
        spans = [item.payload for item in items]
        assert len(spans) == 0  # No client span created without an active span.
    else:
        events = capture_events()

        try:
            async with grpc.aio.insecure_channel(
                f"localhost:{port}"  # noqa: E231
            ) as channel:
                stub = gRPCTestServiceStub(channel)
                with pytest.raises(grpc.RpcError) as exc:
                    await stub.TestServe(gRPCTestMessage(text="test"))
                assert exc.value.details() == "Method not found!"
        finally:
            await server.stop(None)

        assert not events


@pytest.mark.asyncio
@pytest.mark.parametrize("span_streaming", [True, False])
async def test_grpc_server_starts_transaction(
    grpc_server_and_channel,
    capture_events,
    capture_items,
    span_streaming,
):
    _, channel = await grpc_server_and_channel(span_streaming)

    # Use the provided channel
    stub = gRPCTestServiceStub(channel)

    if span_streaming:
        items = capture_items("span")

        await stub.TestServe(gRPCTestMessage(text="test"))

        sentry_sdk.flush()
        spans = [item.payload for item in items]
        span = spans[0]

        assert spans[1]["attributes"]["sentry.segment.name.source"] == "custom"
        assert spans[1]["attributes"]["sentry.op"] == OP.GRPC_SERVER
        assert span["attributes"]["sentry.op"] == "test"
    else:
        events = capture_events()

        await stub.TestServe(gRPCTestMessage(text="test"))

        (event,) = events
        span = event["spans"][0]

        assert event["type"] == "transaction"
        assert event["transaction_info"] == {
            "source": "custom",
        }
        assert event["contexts"]["trace"]["op"] == OP.GRPC_SERVER
        assert span["op"] == "test"


@pytest.mark.asyncio
@pytest.mark.parametrize("span_streaming", [True, False])
async def test_grpc_server_continues_transaction(
    grpc_server_and_channel,
    capture_events,
    capture_items,
    span_streaming,
):
    _, channel = await grpc_server_and_channel(span_streaming)

    # Use the provided channel
    stub = gRPCTestServiceStub(channel)

    if span_streaming:
        items = capture_items("span")

        with sentry_sdk.traces.start_span(name="custom parent") as segment_span:
            metadata = (
                (
                    "baggage",
                    "sentry-trace_id={trace_id},sentry-environment=test,"
                    "sentry-transaction=test-transaction,sentry-sample_rate=1.0".format(
                        trace_id=segment_span.trace_id
                    ),
                ),
                (
                    "sentry-trace",
                    "{trace_id}-{parent_span_id}-{sampled}".format(
                        trace_id=segment_span.trace_id,
                        parent_span_id=segment_span.span_id,
                        sampled=1,
                    ),
                ),
            )

            await stub.TestServe(gRPCTestMessage(text="test"), metadata=metadata)

        sentry_sdk.flush()
        spans = [item.payload for item in items]
        span = spans[0]

        assert spans[1]["attributes"]["sentry.segment.name.source"] == "custom"
        assert spans[1]["attributes"]["sentry.op"] == OP.GRPC_SERVER
        assert spans[1]["trace_id"] == segment_span.trace_id
        assert span["attributes"]["sentry.op"] == "test"
    else:
        events = capture_events()

        with sentry_sdk.start_transaction() as transaction:
            metadata = (
                (
                    "baggage",
                    "sentry-trace_id={trace_id},sentry-environment=test,"
                    "sentry-transaction=test-transaction,sentry-sample_rate=1.0".format(
                        trace_id=transaction.trace_id
                    ),
                ),
                (
                    "sentry-trace",
                    "{trace_id}-{parent_span_id}-{sampled}".format(
                        trace_id=transaction.trace_id,
                        parent_span_id=transaction.span_id,
                        sampled=1,
                    ),
                ),
            )

            await stub.TestServe(gRPCTestMessage(text="test"), metadata=metadata)

        (event, _) = events
        span = event["spans"][0]

        assert event["type"] == "transaction"
        assert event["transaction_info"] == {
            "source": "custom",
        }
        assert event["contexts"]["trace"]["op"] == OP.GRPC_SERVER
        assert event["contexts"]["trace"]["trace_id"] == transaction.trace_id
        assert span["op"] == "test"


@pytest.mark.asyncio
@pytest.mark.parametrize("span_streaming", [True, False])
async def test_grpc_server_exception(
    grpc_server_and_channel,
    capture_events,
    capture_items,
    span_streaming,
):
    _, channel = await grpc_server_and_channel(span_streaming)

    # Use the provided channel
    stub = gRPCTestServiceStub(channel)

    if span_streaming:
        items = capture_items("event")

        try:
            await stub.TestServe(gRPCTestMessage(text="exception"))
            raise AssertionError()
        except Exception:
            pass

        (event,) = (item.payload for item in items if item.type == "event")
    else:
        events = capture_events()

        try:
            await stub.TestServe(gRPCTestMessage(text="exception"))
            raise AssertionError()
        except Exception:
            pass

        (event, _) = events

    assert event["exception"]["values"][0]["type"] == "TestService.TestException"
    assert event["exception"]["values"][0]["value"] == "test"
    assert event["exception"]["values"][0]["mechanism"]["handled"] is False
    assert event["exception"]["values"][0]["mechanism"]["type"] == "grpc"


@pytest.mark.asyncio
async def test_grpc_server_abort(grpc_server_and_channel, capture_events):
    _, channel = await grpc_server_and_channel(False)
    events = capture_events()

    # Use the provided channel
    stub = gRPCTestServiceStub(channel)
    try:
        await stub.TestServe(gRPCTestMessage(text="abort"))
        raise AssertionError()
    except Exception:
        pass

    # Add a small delay to allow events to be collected
    await asyncio.sleep(0.1)

    assert len(events) == 1


@pytest.mark.asyncio
@pytest.mark.parametrize("span_streaming", [True, False])
async def test_grpc_client_starts_span(
    grpc_server_and_channel,
    capture_events_forksafe,
    capture_items_forksafe,
    span_streaming,
):
    _, channel = await grpc_server_and_channel(span_streaming)

    # Use the provided channel
    stub = gRPCTestServiceStub(channel)

    if span_streaming:
        items = capture_items_forksafe("span")

        with sentry_sdk.traces.start_span(name="custom parent") as span:
            await stub.TestServe(gRPCTestMessage(text="test"))

        sentry_sdk.flush()
        items.write_file.close()
        items = items.read_event()
        spans = [item["payload"] for item in items if item["type"] == "span"]
        span = spans[2]

        assert len(spans) == 4
        assert span["attributes"]["sentry.op"] == OP.GRPC_CLIENT
        assert (
            span["name"]
            == "unary unary call to /grpc_test_server.gRPCTestService/TestServe"
        )
        assert span["attributes"] == ApproxDict(
            {
                "rpc.method": "/grpc_test_server.gRPCTestService/TestServe",
                "sentry.environment": mock.ANY,
                "sentry.op": "grpc.client",
                "sentry.origin": "auto.grpc.grpc",
                "sentry.release": mock.ANY,
                "sentry.sdk.name": "sentry.python",
                "sentry.sdk.version": mock.ANY,
                "sentry.segment.id": mock.ANY,
                "sentry.segment.name": "custom parent",
                "server.address": mock.ANY,
                "thread.id": mock.ANY,
                "thread.name": mock.ANY,
                "rpc.response.status_code": "OK",
            }
        )
    else:
        events = capture_events_forksafe()

        with start_transaction():
            await stub.TestServe(gRPCTestMessage(text="test"))

        events.write_file.close()
        events.read_event()
        local_transaction = events.read_event()
        span = local_transaction["spans"][0]

        assert len(local_transaction["spans"]) == 1
        assert span["op"] == OP.GRPC_CLIENT
        assert (
            span["description"]
            == "unary unary call to /grpc_test_server.gRPCTestService/TestServe"
        )
        assert span["data"] == ApproxDict(
            {
                "type": "unary unary",
                "method": "/grpc_test_server.gRPCTestService/TestServe",
                "code": "OK",
            }
        )


@pytest.mark.asyncio
@pytest.mark.parametrize("span_streaming", [True, False])
async def test_grpc_client_unary_stream_starts_span(
    grpc_server_and_channel,
    capture_events_forksafe,
    capture_items_forksafe,
    span_streaming,
):
    _, channel = await grpc_server_and_channel(span_streaming)

    # Use the provided channel
    stub = gRPCTestServiceStub(channel)

    if span_streaming:
        items = capture_items_forksafe("span")

        with sentry_sdk.traces.start_span(name="custom parent"):
            response = stub.TestUnaryStream(gRPCTestMessage(text="test"))
            [_ async for _ in response]

        sentry_sdk.flush()
        items.write_file.close()
        items = items.read_event()
        spans = [item["payload"] for item in items if item["type"] == "span"]
        span = spans[0]

        assert len(spans) == 2
        assert span["attributes"]["sentry.op"] == OP.GRPC_CLIENT
        assert (
            span["name"]
            == "unary stream call to /grpc_test_server.gRPCTestService/TestUnaryStream"
        )
        assert span["attributes"] == ApproxDict(
            {
                "rpc.method": "/grpc_test_server.gRPCTestService/TestUnaryStream",
                "sentry.environment": mock.ANY,
                "sentry.op": "grpc.client",
                "sentry.origin": "auto.grpc.grpc",
                "sentry.release": mock.ANY,
                "sentry.sdk.name": "sentry.python",
                "sentry.sdk.version": mock.ANY,
                "sentry.segment.id": mock.ANY,
                "sentry.segment.name": "custom parent",
                "server.address": mock.ANY,
                "thread.id": mock.ANY,
                "thread.name": mock.ANY,
            }
        )
    else:
        events = capture_events_forksafe()

        with start_transaction():
            response = stub.TestUnaryStream(gRPCTestMessage(text="test"))
            [_ async for _ in response]

        events.write_file.close()
        local_transaction = events.read_event()
        span = local_transaction["spans"][0]

        assert len(local_transaction["spans"]) == 1
        assert span["op"] == OP.GRPC_CLIENT
        assert (
            span["description"]
            == "unary stream call to /grpc_test_server.gRPCTestService/TestUnaryStream"
        )
        assert span["data"] == ApproxDict(
            {
                "type": "unary stream",
                "method": "/grpc_test_server.gRPCTestService/TestUnaryStream",
            }
        )


@pytest.mark.asyncio
async def test_stream_stream(grpc_server_and_channel):
    """
    Test to verify stream-stream works.
    Tracing not supported for it yet.
    """
    _, channel = await grpc_server_and_channel(False)

    # Use the provided channel
    stub = gRPCTestServiceStub(channel)
    response = stub.TestStreamStream((gRPCTestMessage(text="test"),))
    async for r in response:
        assert r.text == "test"


@pytest.mark.asyncio
async def test_stream_unary(grpc_server_and_channel):
    """
    Test to verify stream-stream works.
    Tracing not supported for it yet.
    """
    _, channel = await grpc_server_and_channel(False)

    # Use the provided channel
    stub = gRPCTestServiceStub(channel)
    response = await stub.TestStreamUnary((gRPCTestMessage(text="test"),))
    assert response.text == "test"


@pytest.mark.asyncio
@pytest.mark.parametrize("span_streaming", [True, False])
async def test_span_origin(
    grpc_server_and_channel,
    capture_events_forksafe,
    capture_items_forksafe,
    span_streaming,
):
    _, channel = await grpc_server_and_channel(span_streaming)

    # Use the provided channel
    stub = gRPCTestServiceStub(channel)

    if span_streaming:
        items = capture_items_forksafe("span")

        with sentry_sdk.traces.start_span(name="custom parent"):
            await stub.TestServe(gRPCTestMessage(text="test"))

        sentry_sdk.flush()
        items.write_file.close()
        items = items.read_event()

        spans = [item["payload"] for item in items if item["type"] == "span"]

        assert spans[1]["attributes"]["sentry.origin"] == "auto.grpc.grpc"
        assert (
            spans[0]["attributes"]["sentry.origin"] == "auto.grpc.grpc.TestService.aio"
        )  # manually created in TestService, not the instrumentation

        assert spans[3]["attributes"]["sentry.origin"] == "manual"
        assert spans[2]["attributes"]["sentry.origin"] == "auto.grpc.grpc"
    else:
        events = capture_events_forksafe()

        with start_transaction(name="custom_transaction"):
            await stub.TestServe(gRPCTestMessage(text="test"))

        events.write_file.close()

        transaction_from_integration = events.read_event()
        custom_transaction = events.read_event()

        assert (
            transaction_from_integration["contexts"]["trace"]["origin"]
            == "auto.grpc.grpc"
        )
        assert (
            transaction_from_integration["spans"][0]["origin"]
            == "auto.grpc.grpc.TestService.aio"
        )  # manually created in TestService, not the instrumentation

        assert custom_transaction["contexts"]["trace"]["origin"] == "manual"
        assert custom_transaction["spans"][0]["origin"] == "auto.grpc.grpc"


class TestService(gRPCTestServiceServicer):
    class TestException(Exception):
        __test__ = False

        def __init__(self):
            super().__init__("test")

    @classmethod
    async def TestServe(cls, request, context):  # noqa: N802
        client = sentry_sdk.get_client()
        span_streaming = has_span_streaming_enabled(client.options)

        if span_streaming:
            with sentry_sdk.traces.start_span(
                name="test",
                attributes={
                    "sentry.op": "test",
                    "sentry.origin": "auto.grpc.grpc.TestService.aio",
                },
            ):
                pass
        else:
            with start_span(
                op="test",
                name="test",
                origin="auto.grpc.grpc.TestService.aio",
            ):
                pass

        if request.text == "exception":
            raise cls.TestException()

        if request.text == "abort":
            await context.abort(grpc.StatusCode.ABORTED, "Aborted!")

        return gRPCTestMessage(text=request.text)

    @classmethod
    async def TestUnaryStream(cls, request, context):  # noqa: N802
        for _ in range(3):
            yield gRPCTestMessage(text=request.text)

    @classmethod
    async def TestStreamStream(cls, request, context):  # noqa: N802
        async for r in request:
            yield r

    @classmethod
    async def TestStreamUnary(cls, request, context):  # noqa: N802
        requests = [r async for r in request]
        return requests.pop()



import itertools
from copy import copy
from unittest import mock

import pytest

import sentry_sdk
from sentry_sdk.integrations.celery import _update_celery_task_headers
from sentry_sdk.tracing_utils import Baggage

BAGGAGE_VALUE = (
    "sentry-trace_id=771a43a4192642f0b136d5159a501700,"
    "sentry-public_key=49d0f7386ad645858ae85020e393bef3,"
    "sentry-sample_rate=0.1337,"
    "custom=value"
)

SENTRY_TRACE_VALUE = "771a43a4192642f0b136d5159a501700-1234567890abcdef-1"


@pytest.mark.parametrize("monitor_beat_tasks", [True, False, None, "", "bla", 1, 0])
def test_monitor_beat_tasks(monitor_beat_tasks):
    headers = {}
    span = None

    outgoing_headers = _update_celery_task_headers(headers, span, monitor_beat_tasks)

    assert headers == {}  # left unchanged

    if monitor_beat_tasks:
        assert outgoing_headers["sentry-monitor-start-timestamp-s"] == mock.ANY
        assert (
            outgoing_headers["headers"]["sentry-monitor-start-timestamp-s"] == mock.ANY
        )
    else:
        assert "sentry-monitor-start-timestamp-s" not in outgoing_headers
        assert "sentry-monitor-start-timestamp-s" not in outgoing_headers["headers"]


@pytest.mark.parametrize("monitor_beat_tasks", [True, False, None, "", "bla", 1, 0])
def test_monitor_beat_tasks_with_headers(monitor_beat_tasks):
    headers = {
        "blub": "foo",
        "sentry-something": "bar",
        "sentry-task-enqueued-time": mock.ANY,
    }
    span = None

    outgoing_headers = _update_celery_task_headers(headers, span, monitor_beat_tasks)

    assert headers == {
        "blub": "foo",
        "sentry-something": "bar",
        "sentry-task-enqueued-time": mock.ANY,
    }  # left unchanged

    if monitor_beat_tasks:
        assert outgoing_headers["blub"] == "foo"
        assert outgoing_headers["sentry-something"] == "bar"
        assert outgoing_headers["sentry-monitor-start-timestamp-s"] == mock.ANY
        assert outgoing_headers["headers"]["sentry-something"] == "bar"
        assert (
            outgoing_headers["headers"]["sentry-monitor-start-timestamp-s"] == mock.ANY
        )
    else:
        assert outgoing_headers["blub"] == "foo"
        assert outgoing_headers["sentry-something"] == "bar"
        assert "sentry-monitor-start-timestamp-s" not in outgoing_headers
        assert "sentry-monitor-start-timestamp-s" not in outgoing_headers["headers"]


def test_span_with_transaction(sentry_init):
    sentry_init(traces_sample_rate=1.0)
    headers = {}
    monitor_beat_tasks = False

    with sentry_sdk.start_transaction(name="test_transaction") as transaction:
        with sentry_sdk.start_span(op="test_span") as span:
            outgoing_headers = _update_celery_task_headers(
                headers, span, monitor_beat_tasks
            )

            assert outgoing_headers["sentry-trace"] == span.to_traceparent()
            assert outgoing_headers["headers"]["sentry-trace"] == span.to_traceparent()
            assert outgoing_headers["baggage"] == transaction.get_baggage().serialize()
            assert (
                outgoing_headers["headers"]["baggage"]
                == transaction.get_baggage().serialize()
            )


def test_span_with_transaction_custom_headers(sentry_init):
    sentry_init(traces_sample_rate=1.0)
    headers = {
        "baggage": BAGGAGE_VALUE,
        "sentry-trace": SENTRY_TRACE_VALUE,
    }

    with sentry_sdk.start_transaction(name="test_transaction") as transaction:
        with sentry_sdk.start_span(op="test_span") as span:
            outgoing_headers = _update_celery_task_headers(headers, span, False)

            assert outgoing_headers["sentry-trace"] == span.to_traceparent()
            assert outgoing_headers["headers"]["sentry-trace"] == span.to_traceparent()

            incoming_baggage = Baggage.from_incoming_header(headers["baggage"])
            combined_baggage = copy(transaction.get_baggage())
            combined_baggage.sentry_items.update(incoming_baggage.sentry_items)
            combined_baggage.third_party_items = ",".join(
                [
                    x
                    for x in [
                        combined_baggage.third_party_items,
                        incoming_baggage.third_party_items,
                    ]
                    if x is not None and x != ""
                ]
            )
            assert outgoing_headers["baggage"] == combined_baggage.serialize(
                include_third_party=True
            )
            assert outgoing_headers["headers"]["baggage"] == combined_baggage.serialize(
                include_third_party=True
            )


@pytest.mark.parametrize("monitor_beat_tasks", [True, False])
def test_celery_trace_propagation_default(sentry_init, monitor_beat_tasks):
    """
    The celery integration does not check the traces_sample_rate.
    By default traces_sample_rate is None which means "do not propagate traces".
    But the celery integration does not check this value.
    The Celery integration has its own mechanism to propagate traces:
    https://docs.sentry.io/platforms/python/integrations/celery/#distributed-traces
    """
    sentry_init()

    headers = {}
    span = None

    scope = sentry_sdk.get_isolation_scope()

    outgoing_headers = _update_celery_task_headers(headers, span, monitor_beat_tasks)

    assert outgoing_headers["sentry-trace"] == scope.get_traceparent()
    assert outgoing_headers["headers"]["sentry-trace"] == scope.get_traceparent()
    assert outgoing_headers["baggage"] == scope.get_baggage().serialize()
    assert outgoing_headers["headers"]["baggage"] == scope.get_baggage().serialize()

    if monitor_beat_tasks:
        assert "sentry-monitor-start-timestamp-s" in outgoing_headers
        assert "sentry-monitor-start-timestamp-s" in outgoing_headers["headers"]
    else:
        assert "sentry-monitor-start-timestamp-s" not in outgoing_headers
        assert "sentry-monitor-start-timestamp-s" not in outgoing_headers["headers"]


@pytest.mark.parametrize(
    "traces_sample_rate,monitor_beat_tasks",
    list(itertools.product([None, 0, 0.0, 0.5, 1.0, 1, 2], [True, False])),
)
def test_celery_trace_propagation_traces_sample_rate(
    sentry_init, traces_sample_rate, monitor_beat_tasks
):
    """
    The celery integration does not check the traces_sample_rate.
    By default traces_sample_rate is None which means "do not propagate traces".
    But the celery integration does not check this value.
    The Celery integration has its own mechanism to propagate traces:
    https://docs.sentry.io/platforms/python/integrations/celery/#distributed-traces
    """
    sentry_init(traces_sample_rate=traces_sample_rate)

    headers = {}
    span = None

    scope = sentry_sdk.get_isolation_scope()

    outgoing_headers = _update_celery_task_headers(headers, span, monitor_beat_tasks)

    assert outgoing_headers["sentry-trace"] == scope.get_traceparent()
    assert outgoing_headers["headers"]["sentry-trace"] == scope.get_traceparent()
    assert outgoing_headers["baggage"] == scope.get_baggage().serialize()
    assert outgoing_headers["headers"]["baggage"] == scope.get_baggage().serialize()

    if monitor_beat_tasks:
        assert "sentry-monitor-start-timestamp-s" in outgoing_headers
        assert "sentry-monitor-start-timestamp-s" in outgoing_headers["headers"]
    else:
        assert "sentry-monitor-start-timestamp-s" not in outgoing_headers
        assert "sentry-monitor-start-timestamp-s" not in outgoing_headers["headers"]


@pytest.mark.parametrize(
    "enable_tracing,monitor_beat_tasks",
    list(itertools.product([None, True, False], [True, False])),
)
def test_celery_trace_propagation_enable_tracing(
    sentry_init, enable_tracing, monitor_beat_tasks
):
    """
    The celery integration does not check the traces_sample_rate.
    By default traces_sample_rate is None which means "do not propagate traces".
    But the celery integration does not check this value.
    The Celery integration has its own mechanism to propagate traces:
    https://docs.sentry.io/platforms/python/integrations/celery/#distributed-traces
    """
    sentry_init(enable_tracing=enable_tracing)

    headers = {}
    span = None

    scope = sentry_sdk.get_isolation_scope()

    outgoing_headers = _update_celery_task_headers(headers, span, monitor_beat_tasks)

    assert outgoing_headers["sentry-trace"] == scope.get_traceparent()
    assert outgoing_headers["headers"]["sentry-trace"] == scope.get_traceparent()
    assert outgoing_headers["baggage"] == scope.get_baggage().serialize()
    assert outgoing_headers["headers"]["baggage"] == scope.get_baggage().serialize()

    if monitor_beat_tasks:
        assert "sentry-monitor-start-timestamp-s" in outgoing_headers
        assert "sentry-monitor-start-timestamp-s" in outgoing_headers["headers"]
    else:
        assert "sentry-monitor-start-timestamp-s" not in outgoing_headers
        assert "sentry-monitor-start-timestamp-s" not in outgoing_headers["headers"]



import sentry_sdk


def test_span_origin_manual(sentry_init, capture_events):
    sentry_init(traces_sample_rate=1.0)
    events = capture_events()

    with sentry_sdk.start_transaction(name="hi"):
        with sentry_sdk.start_span(op="foo", name="bar"):
            pass

    (event,) = events

    assert len(events) == 1
    assert event["spans"][0]["origin"] == "manual"
    assert event["contexts"]["trace"]["origin"] == "manual"


def test_span_origin_custom(sentry_init, capture_events):
    sentry_init(traces_sample_rate=1.0)
    events = capture_events()

    with sentry_sdk.start_transaction(name="hi"):
        with sentry_sdk.start_span(op="foo", name="bar", origin="foo.foo2.foo3"):
            pass

    with sentry_sdk.start_transaction(name="ho", origin="ho.ho2.ho3"):
        with sentry_sdk.start_span(op="baz", name="qux", origin="baz.baz2.baz3"):
            pass

    (first_transaction, second_transaction) = events

    assert len(events) == 2
    assert first_transaction["contexts"]["trace"]["origin"] == "manual"
    assert first_transaction["spans"][0]["origin"] == "foo.foo2.foo3"

    assert second_transaction["contexts"]["trace"]["origin"] == "ho.ho2.ho3"
    assert second_transaction["spans"][0]["origin"] == "baz.baz2.baz3"


def test_span_origin_manual_span_streaming(sentry_init, capture_items):
    sentry_init(trace_lifecycle="stream", traces_sample_rate=1.0)
    items = capture_items("span")

    with sentry_sdk.traces.start_span(name="hi"):
        pass

    sentry_sdk.flush()

    (span,) = [item.payload for item in items]

    assert len(items) == 1
    assert span["attributes"]["sentry.origin"] == "manual"


def test_span_origin_custom_span_streaming(sentry_init, capture_items):
    sentry_init(trace_lifecycle="stream", traces_sample_rate=1.0)
    items = capture_items("span")

    with sentry_sdk.traces.start_span(
        name="hi", attributes={"sentry.origin": "foo.foo2.foo3"}
    ):
        pass

    with sentry_sdk.traces.start_span(
        name="ho", attributes={"sentry.origin": "baz.baz2.baz3"}
    ):
        pass

    sentry_sdk.flush()

    (span1, span2) = [item.payload for item in items]

    assert len(items) == 2
    assert span1["attributes"]["sentry.origin"] == "foo.foo2.foo3"
    assert span2["attributes"]["sentry.origin"] == "baz.baz2.baz3"



import os
import sys

import pytest

pytest.importorskip("sqlalchemy")

# Load `sqlalchemy_helpers` into the module search path to test query source path names relative to module. See
# `test_query_source_with_module_in_search_path`
sys.path.insert(0, os.path.join(os.path.dirname(__file__)))



from ariadne import ObjectType, QueryType, gql, graphql_sync, make_executable_schema
from ariadne.asgi import GraphQL
from fastapi import FastAPI
from fastapi.testclient import TestClient
from flask import Flask, jsonify, request

from sentry_sdk.integrations.ariadne import AriadneIntegration
from sentry_sdk.integrations.fastapi import FastApiIntegration
from sentry_sdk.integrations.flask import FlaskIntegration
from sentry_sdk.integrations.starlette import StarletteIntegration


def schema_factory():
    type_defs = gql(
        """
        type Query {
            greeting(name: String): Greeting
            error: String
        }

        type Greeting {
            name: String
        }
    """
    )

    query = QueryType()
    greeting = ObjectType("Greeting")

    @query.field("greeting")
    def resolve_greeting(*_, **kwargs):
        name = kwargs.pop("name")
        return {"name": name}

    @query.field("error")
    def resolve_error(obj, *_):
        raise RuntimeError("resolver failed")

    @greeting.field("name")
    def resolve_name(obj, *_):
        return "Hello, {}!".format(obj["name"])

    return make_executable_schema(type_defs, query)


def test_capture_request_and_response_if_send_pii_is_on_async(
    sentry_init, capture_events
):
    sentry_init(
        send_default_pii=True,
        integrations=[
            AriadneIntegration(),
            FastApiIntegration(),
            StarletteIntegration(),
        ],
    )
    events = capture_events()

    schema = schema_factory()

    async_app = FastAPI()
    async_app.mount("/graphql/", GraphQL(schema))

    query = {"query": "query ErrorQuery {error}"}
    client = TestClient(async_app)
    client.post("/graphql", json=query)

    assert len(events) == 1

    (event,) = events
    assert event["exception"]["values"][0]["mechanism"]["type"] == "ariadne"
    assert event["contexts"]["response"] == {
        "data": {
            "data": {"error": None},
            "errors": [
                {
                    "locations": [{"column": 19, "line": 1}],
                    "message": "resolver failed",
                    "path": ["error"],
                }
            ],
        }
    }
    assert event["request"]["api_target"] == "graphql"
    assert event["request"]["data"] == query


def test_capture_request_and_response_if_send_pii_is_on_sync(
    sentry_init, capture_events
):
    sentry_init(
        send_default_pii=True,
        integrations=[AriadneIntegration(), FlaskIntegration()],
    )
    events = capture_events()

    schema = schema_factory()

    sync_app = Flask(__name__)

    @sync_app.route("/graphql", methods=["POST"])
    def graphql_server():
        data = request.get_json()
        success, result = graphql_sync(schema, data)
        return jsonify(result), 200

    query = {"query": "query ErrorQuery {error}"}
    client = sync_app.test_client()
    client.post("/graphql", json=query)

    assert len(events) == 1

    (event,) = events
    assert event["exception"]["values"][0]["mechanism"]["type"] == "ariadne"
    assert event["contexts"]["response"] == {
        "data": {
            "data": {"error": None},
            "errors": [
                {
                    "locations": [{"column": 19, "line": 1}],
                    "message": "resolver failed",
                    "path": ["error"],
                }
            ],
        }
    }
    assert event["request"]["api_target"] == "graphql"
    assert event["request"]["data"] == query


def test_do_not_capture_request_and_response_if_send_pii_is_off_async(
    sentry_init, capture_events
):
    sentry_init(
        integrations=[
            AriadneIntegration(),
            FastApiIntegration(),
            StarletteIntegration(),
        ],
    )
    events = capture_events()

    schema = schema_factory()

    async_app = FastAPI()
    async_app.mount("/graphql/", GraphQL(schema))

    query = {"query": "query ErrorQuery {error}"}
    client = TestClient(async_app)
    client.post("/graphql", json=query)

    assert len(events) == 1

    (event,) = events
    assert event["exception"]["values"][0]["mechanism"]["type"] == "ariadne"
    assert "data" not in event["request"]
    assert "response" not in event["contexts"]


def test_do_not_capture_request_and_response_if_send_pii_is_off_sync(
    sentry_init, capture_events
):
    sentry_init(
        integrations=[AriadneIntegration(), FlaskIntegration()],
    )
    events = capture_events()

    schema = schema_factory()

    sync_app = Flask(__name__)

    @sync_app.route("/graphql", methods=["POST"])
    def graphql_server():
        data = request.get_json()
        success, result = graphql_sync(schema, data)
        return jsonify(result), 200

    query = {"query": "query ErrorQuery {error}"}
    client = sync_app.test_client()
    client.post("/graphql", json=query)

    assert len(events) == 1

    (event,) = events
    assert event["exception"]["values"][0]["mechanism"]["type"] == "ariadne"
    assert "data" not in event["request"]
    assert "response" not in event["contexts"]


def test_capture_validation_error(sentry_init, capture_events):
    sentry_init(
        send_default_pii=True,
        integrations=[
            AriadneIntegration(),
            FastApiIntegration(),
            StarletteIntegration(),
        ],
    )
    events = capture_events()

    schema = schema_factory()

    async_app = FastAPI()
    async_app.mount("/graphql/", GraphQL(schema))

    query = {"query": "query ErrorQuery {doesnt_exist}"}
    client = TestClient(async_app)
    client.post("/graphql", json=query)

    assert len(events) == 1

    (event,) = events
    assert event["exception"]["values"][0]["mechanism"]["type"] == "ariadne"
    assert event["contexts"]["response"] == {
        "data": {
            "errors": [
                {
                    "locations": [{"column": 19, "line": 1}],
                    "message": "Cannot query field 'doesnt_exist' on type 'Query'.",
                }
            ]
        }
    }
    assert event["request"]["api_target"] == "graphql"
    assert event["request"]["data"] == query


def test_no_event_if_no_errors_async(sentry_init, capture_events):
    sentry_init(
        integrations=[
            AriadneIntegration(),
            FastApiIntegration(),
            StarletteIntegration(),
        ],
    )
    events = capture_events()

    schema = schema_factory()

    async_app = FastAPI()
    async_app.mount("/graphql/", GraphQL(schema))

    query = {
        "query": "query GreetingQuery($name: String) { greeting(name: $name) {name} }",
        "variables": {"name": "some name"},
    }
    client = TestClient(async_app)
    client.post("/graphql", json=query)

    assert len(events) == 0


def test_no_event_if_no_errors_sync(sentry_init, capture_events):
    sentry_init(
        integrations=[AriadneIntegration(), FlaskIntegration()],
    )
    events = capture_events()

    schema = schema_factory()

    sync_app = Flask(__name__)

    @sync_app.route("/graphql", methods=["POST"])
    def graphql_server():
        data = request.get_json()
        success, result = graphql_sync(schema, data)
        return jsonify(result), 200

    query = {
        "query": "query GreetingQuery($name: String) { greeting(name: $name) {name} }",
        "variables": {"name": "some name"},
    }
    client = sync_app.test_client()
    client.post("/graphql", json=query)

    assert len(events) == 0



import re
from unittest.mock import ANY, MagicMock, patch

import pytest
from loguru import logger
from loguru._recattrs import RecordFile, RecordLevel

import sentry_sdk
from sentry_sdk.consts import VERSION
from sentry_sdk.integrations.loguru import LoggingLevels, LoguruIntegration

logger.remove(0)  # don't print to console


@pytest.mark.parametrize(
    "level,created_event,expected_sentry_level",
    [
        # None - no breadcrumb
        # False - no event
        # True - event created
        (LoggingLevels.TRACE, None, "debug"),
        (LoggingLevels.DEBUG, None, "debug"),
        (LoggingLevels.INFO, False, "info"),
        (LoggingLevels.SUCCESS, False, "info"),
        (LoggingLevels.WARNING, False, "warning"),
        (LoggingLevels.ERROR, True, "error"),
        (LoggingLevels.CRITICAL, True, "critical"),
    ],
)
@pytest.mark.parametrize("disable_breadcrumbs", [True, False])
@pytest.mark.parametrize("disable_events", [True, False])
def test_just_log(
    sentry_init,
    capture_events,
    level,
    created_event,
    expected_sentry_level,
    disable_breadcrumbs,
    disable_events,
    uninstall_integration,
    request,
):
    uninstall_integration("loguru")
    request.addfinalizer(logger.remove)

    sentry_init(
        integrations=[
            LoguruIntegration(
                level=None if disable_breadcrumbs else LoggingLevels.INFO.value,
                event_level=None if disable_events else LoggingLevels.ERROR.value,
            )
        ],
        default_integrations=False,
    )
    events = capture_events()

    getattr(logger, level.name.lower())("test")

    expected_pattern = (
        r" \| "
        + r"{:9}".format(level.name.upper())
        + r"\| tests\.integrations\.loguru\.test_loguru:test_just_log:\d+ - test"
    )

    if not created_event:
        assert not events

        breadcrumbs = sentry_sdk.get_isolation_scope()._breadcrumbs
        if (
            not disable_breadcrumbs and created_event is not None
        ):  # not None == not TRACE or DEBUG level
            (breadcrumb,) = breadcrumbs
            assert breadcrumb["level"] == expected_sentry_level
            assert breadcrumb["category"] == "tests.integrations.loguru.test_loguru"
            assert re.fullmatch(expected_pattern, breadcrumb["message"][23:])
        else:
            assert not breadcrumbs

        return

    if disable_events:
        assert not events
        return

    (event,) = events
    assert event["level"] == expected_sentry_level
    assert event["logger"] == "tests.integrations.loguru.test_loguru"
    assert re.fullmatch(expected_pattern, event["logentry"]["message"][23:])


def test_breadcrumb_format(sentry_init, capture_events, uninstall_integration, request):
    uninstall_integration("loguru")
    request.addfinalizer(logger.remove)

    sentry_init(
        integrations=[
            LoguruIntegration(
                level=LoggingLevels.INFO.value,
                event_level=None,
                breadcrumb_format="{message}",
            )
        ],
        default_integrations=False,
    )

    logger.info("test")
    formatted_message = "test"

    breadcrumbs = sentry_sdk.get_isolation_scope()._breadcrumbs
    (breadcrumb,) = breadcrumbs
    assert breadcrumb["message"] == formatted_message


def test_event_format(sentry_init, capture_events, uninstall_integration, request):
    uninstall_integration("loguru")
    request.addfinalizer(logger.remove)

    sentry_init(
        integrations=[
            LoguruIntegration(
                level=None,
                event_level=LoggingLevels.ERROR.value,
                event_format="{message}",
            )
        ],
        default_integrations=False,
    )
    events = capture_events()

    logger.error("test")
    formatted_message = "test"

    (event,) = events
    assert event["logentry"]["message"] == formatted_message


def test_sentry_logs_warning(
    sentry_init, capture_items, uninstall_integration, request
):
    uninstall_integration("loguru")
    request.addfinalizer(logger.remove)

    sentry_init(enable_logs=True)
    items = capture_items("log")

    logger.warning("this is {} a {}", "just", "template")

    sentry_sdk.get_client().flush()
    logs = [item.payload for item in items]

    attrs = logs[0]["attributes"]
    assert "code.file.path" in attrs
    assert "code.line.number" in attrs
    assert attrs["logger.name"] == "tests.integrations.loguru.test_loguru"
    assert attrs["sentry.environment"] == "production"
    assert attrs["sentry.origin"] == "auto.log.loguru"
    assert logs[0]["attributes"]["sentry.severity_number"] == 13
    assert logs[0]["attributes"]["sentry.severity_text"] == "warn"


def test_sentry_logs_debug(
    sentry_init, capture_envelopes, uninstall_integration, request
):
    uninstall_integration("loguru")
    request.addfinalizer(logger.remove)

    sentry_init(enable_logs=True)
    envelopes = capture_envelopes()

    logger.debug("this is %s a template %s", "1", "2")
    sentry_sdk.get_client().flush()

    assert len(envelopes) == 0


def test_sentry_log_levels(sentry_init, capture_items, uninstall_integration, request):
    uninstall_integration("loguru")
    request.addfinalizer(logger.remove)

    sentry_init(
        integrations=[LoguruIntegration(sentry_logs_level=LoggingLevels.SUCCESS)],
        enable_logs=True,
    )
    items = capture_items("log")

    logger.trace("this is a log")
    logger.debug("this is a log")
    logger.info("this is a log")
    logger.success("this is a log")
    logger.warning("this is a log")
    logger.error("this is a log")
    logger.critical("this is a log")

    sentry_sdk.get_client().flush()
    logs = [item.payload for item in items]
    assert len(logs) == 4

    assert logs[0]["attributes"]["sentry.severity_number"] == 11
    assert logs[0]["attributes"]["sentry.severity_text"] == "info"
    assert logs[1]["attributes"]["sentry.severity_number"] == 13
    assert logs[1]["attributes"]["sentry.severity_text"] == "warn"
    assert logs[2]["attributes"]["sentry.severity_number"] == 17
    assert logs[2]["attributes"]["sentry.severity_text"] == "error"
    assert logs[3]["attributes"]["sentry.severity_number"] == 21
    assert logs[3]["attributes"]["sentry.severity_text"] == "fatal"


def test_disable_loguru_logs(
    sentry_init, capture_items, uninstall_integration, request
):
    uninstall_integration("loguru")
    request.addfinalizer(logger.remove)

    sentry_init(
        integrations=[LoguruIntegration(sentry_logs_level=None)],
        enable_logs=True,
    )
    items = capture_items("log")

    logger.trace("this is a log")
    logger.debug("this is a log")
    logger.info("this is a log")
    logger.success("this is a log")
    logger.warning("this is a log")
    logger.error("this is a log")
    logger.critical("this is a log")

    sentry_sdk.get_client().flush()
    logs = [item.payload for item in items]
    assert len(logs) == 0


def test_disable_sentry_logs(
    sentry_init, capture_items, uninstall_integration, request
):
    uninstall_integration("loguru")
    request.addfinalizer(logger.remove)

    sentry_init(
        _experiments={"enable_logs": False},
    )
    items = capture_items("log")

    logger.trace("this is a log")
    logger.debug("this is a log")
    logger.info("this is a log")
    logger.success("this is a log")
    logger.warning("this is a log")
    logger.error("this is a log")
    logger.critical("this is a log")

    sentry_sdk.get_client().flush()
    logs = [item.payload for item in items]
    assert len(logs) == 0


def test_no_log_infinite_loop(
    sentry_init, capture_envelopes, uninstall_integration, request
):
    """
    In debug mode, there should be no infinite loops even when a low log level is set.
    """
    uninstall_integration("loguru")
    request.addfinalizer(logger.remove)

    sentry_init(
        enable_logs=True,
        integrations=[LoguruIntegration(sentry_logs_level=LoggingLevels.DEBUG)],
        debug=True,
    )
    envelopes = capture_envelopes()

    logger.debug("this is %s a template %s", "1", "2")
    sentry_sdk.get_client().flush()

    assert len(envelopes) == 1


def test_logging_errors(
    sentry_init, capture_envelopes, capture_items, uninstall_integration, request
):
    """We're able to log errors without erroring."""
    uninstall_integration("loguru")
    request.addfinalizer(logger.remove)

    sentry_init(enable_logs=True)
    envelopes = capture_envelopes()
    items = capture_items("log")

    logger.error(Exception("test exc 1"))
    logger.error("error is %s", Exception("test exc 2"))
    sentry_sdk.get_client().flush()

    error_event_1 = envelopes[0].items[0].payload.json
    assert error_event_1["level"] == "error"
    error_event_2 = envelopes[1].items[0].payload.json
    assert error_event_2["level"] == "error"

    logs = [item.payload for item in items]
    assert logs[0]["attributes"]["sentry.severity_text"] == "error"
    assert "code.line.number" in logs[0]["attributes"]

    assert logs[1]["attributes"]["sentry.severity_text"] == "error"
    assert "code.line.number" in logs[1]["attributes"]

    assert len(logs) == 2


def test_log_strips_project_root(
    sentry_init, capture_items, uninstall_integration, request
):
    uninstall_integration("loguru")
    request.addfinalizer(logger.remove)

    sentry_init(
        enable_logs=True,
        project_root="/custom/test",
    )
    items = capture_items("log")

    class FakeMessage:
        def __init__(self, *args, **kwargs):
            pass

        @property
        def record(self):
            return {
                "elapsed": MagicMock(),
                "exception": None,
                "file": RecordFile(name="app.py", path="/custom/test/blah/path.py"),
                "function": "<module>",
                "level": RecordLevel(name="ERROR", no=20, icon=""),
                "line": 35,
                "message": "some message",
                "module": "app",
                "name": "__main__",
                "process": MagicMock(),
                "thread": MagicMock(),
                "time": MagicMock(),
                "extra": MagicMock(),
            }

        @record.setter
        def record(self, val):
            pass

    with patch("loguru._handler.Message", FakeMessage):
        logger.error("some message")

    sentry_sdk.get_client().flush()

    logs = [item.payload for item in items]
    assert len(logs) == 1
    attrs = logs[0]["attributes"]
    assert attrs["code.file.path"] == "blah/path.py"


def test_log_keeps_full_path_if_not_in_project_root(
    sentry_init, capture_items, uninstall_integration, request
):
    uninstall_integration("loguru")
    request.addfinalizer(logger.remove)

    sentry_init(
        enable_logs=True,
        project_root="/custom/test",
    )
    items = capture_items("log")

    class FakeMessage:
        def __init__(self, *args, **kwargs):
            pass

        @property
        def record(self):
            return {
                "elapsed": MagicMock(),
                "exception": None,
                "file": RecordFile(name="app.py", path="/blah/path.py"),
                "function": "<module>",
                "level": RecordLevel(name="ERROR", no=20, icon=""),
                "line": 35,
                "message": "some message",
                "module": "app",
                "name": "__main__",
                "process": MagicMock(),
                "thread": MagicMock(),
                "time": MagicMock(),
                "extra": MagicMock(),
            }

        @record.setter
        def record(self, val):
            pass

    with patch("loguru._handler.Message", FakeMessage):
        logger.error("some message")

    sentry_sdk.get_client().flush()

    logs = [item.payload for item in items]
    assert len(logs) == 1
    attrs = logs[0]["attributes"]
    assert attrs["code.file.path"] == "/blah/path.py"


def test_logger_with_all_attributes(
    sentry_init, capture_items, uninstall_integration, request
):
    uninstall_integration("loguru")
    request.addfinalizer(logger.remove)

    sentry_init(enable_logs=True)
    items = capture_items("log")

    logger.warning("log #{}", 1)
    sentry_sdk.get_client().flush()

    logs = [item.payload for item in items]

    assert "span_id" not in logs[0]

    attributes = logs[0]["attributes"]

    assert "process.pid" in attributes
    assert isinstance(attributes["process.pid"], int)
    del attributes["process.pid"]

    assert "sentry.release" in attributes
    assert isinstance(attributes["sentry.release"], str)
    del attributes["sentry.release"]

    assert "server.address" in attributes
    assert isinstance(attributes["server.address"], str)
    del attributes["server.address"]

    assert "thread.id" in attributes
    assert isinstance(attributes["thread.id"], int)
    del attributes["thread.id"]

    assert "code.file.path" in attributes
    assert isinstance(attributes["code.file.path"], str)
    del attributes["code.file.path"]

    assert "code.function.name" in attributes
    assert isinstance(attributes["code.function.name"], str)
    del attributes["code.function.name"]

    assert "code.line.number" in attributes
    assert isinstance(attributes["code.line.number"], int)
    del attributes["code.line.number"]

    assert "process.executable.name" in attributes
    assert isinstance(attributes["process.executable.name"], str)
    del attributes["process.executable.name"]

    assert "thread.name" in attributes
    assert isinstance(attributes["thread.name"], str)
    del attributes["thread.name"]

    assert attributes.pop("sentry.sdk.name").startswith("sentry.python")

    # Assert on the remaining non-dynamic attributes.
    assert attributes == {
        "logger.name": "tests.integrations.loguru.test_loguru",
        "sentry.origin": "auto.log.loguru",
        "sentry.environment": "production",
        "process.runtime.name": ANY,
        "process.runtime.version": ANY,
        "sentry.sdk.version": VERSION,
        "sentry.severity_number": 13,
        "sentry.severity_text": "warn",
    }


def test_logger_capture_parameters_from_args(
    sentry_init, capture_items, uninstall_integration, request
):
    # This is currently not supported as regular args don't get added to extra
    # (which we use for populating parameters). Adding this test to make that
    # explicit and so that it's easy to change later.
    uninstall_integration("loguru")
    request.addfinalizer(logger.remove)

    sentry_init(enable_logs=True)
    items = capture_items("log")

    logger.warning("Task ID: {}", 123)

    sentry_sdk.get_client().flush()

    logs = [item.payload for item in items]

    attributes = logs[0]["attributes"]
    assert "sentry.message.parameter.0" not in attributes


def test_logger_capture_parameters_from_kwargs(
    sentry_init, capture_items, uninstall_integration, request
):
    uninstall_integration("loguru")
    request.addfinalizer(logger.remove)

    sentry_init(enable_logs=True)
    items = capture_items("log")

    logger.warning("Task ID: {task_id}", task_id=123)

    sentry_sdk.get_client().flush()

    logs = [item.payload for item in items]

    attributes = logs[0]["attributes"]
    assert attributes["sentry.message.parameter.task_id"] == 123


def test_logger_capture_parameters_from_contextualize(
    sentry_init, capture_items, uninstall_integration, request
):
    uninstall_integration("loguru")
    request.addfinalizer(logger.remove)

    sentry_init(enable_logs=True)
    items = capture_items("log")

    with logger.contextualize(task_id=123):
        logger.warning("Log")

    sentry_sdk.get_client().flush()

    logs = [item.payload for item in items]

    attributes = logs[0]["attributes"]
    assert attributes["sentry.message.parameter.task_id"] == 123


def test_logger_capture_parameters_from_bind(
    sentry_init, capture_items, uninstall_integration, request
):
    uninstall_integration("loguru")
    request.addfinalizer(logger.remove)

    sentry_init(enable_logs=True)
    items = capture_items("log")

    logger.bind(task_id=123).warning("Log")
    sentry_sdk.get_client().flush()

    logs = [item.payload for item in items]

    attributes = logs[0]["attributes"]
    assert attributes["sentry.message.parameter.task_id"] == 123


def test_logger_capture_parameters_from_patch(
    sentry_init, capture_items, uninstall_integration, request
):
    uninstall_integration("loguru")
    request.addfinalizer(logger.remove)

    sentry_init(enable_logs=True)
    items = capture_items("log")

    logger.patch(lambda record: record["extra"].update(task_id=123)).warning("Log")
    sentry_sdk.get_client().flush()

    logs = [item.payload for item in items]

    attributes = logs[0]["attributes"]
    assert attributes["sentry.message.parameter.task_id"] == 123


def test_no_parameters_no_template(
    sentry_init, capture_items, uninstall_integration, request
):
    uninstall_integration("loguru")
    request.addfinalizer(logger.remove)

    sentry_init(enable_logs=True)
    items = capture_items("log")

    logger.warning("Logging a hardcoded warning")
    sentry_sdk.get_client().flush()

    logs = [item.payload for item in items]

    attributes = logs[0]["attributes"]
    assert "sentry.message.template" not in attributes
